"""In this file, we analyse the action library generated by Teacher1 and Teacher2.

Through

    1. filtering out actions that decrease rho less effectively
    2. filtering out actions that occur less frequently
    3. filtering out "do nothing" action
    4. add your filtering rules...,

We obtain an action space in the form of numpy array.

Credit: The methods are the enhanced methods of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""
import logging
from pathlib import Path
from typing import Union, Collection, List, Optional

import defopt
import grid2op
import numpy as np
import pandas as pd
from grid2op.Action import BaseAction
from grid2op.Environment import BaseEnv
from matplotlib import pyplot as plt

from curriculumagent.teacher.submodule.common import affected_substations
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction


def read_experience(experience_files: Union[Path, Collection[Path]]) -> pd.DataFrame:
    """Read teacher_experience data from csv file and concat it if needed.

    Args:
        experience_files:Either one file or multiple files with teacher_experience data.

    Returns:
        Dataframe with teacher_experience.

    """
    data = None
    for exp_file in experience_files:
        if data is None:
            data = pd.read_csv(exp_file, on_bad_lines="skip")
        else:
            data = pd.concat((data, pd.read_csv(exp_file, on_bad_lines="skip")))
    return data


def filter_good_experience(data: pd.DataFrame) -> pd.DataFrame:
    """Filter actions based on criteria.

    The actions are either filtered if:
        1. The action decreases the rho by less than 2%
        2. The action does nothing

    Args:
        data: Input of teacher_experience.

    Returns:
        Reduced dataframe with filtered action/teacher_experience.

    """
    good_samples = data["rho_improvement"] > 0.02
    return data[good_samples]


def rank_actions_simple(
        data: pd.DataFrame, env: BaseEnv, best_n: int = 416, plot_choice: Optional[bool] = False
) -> List[BaseAction]:
    """Rank the given teacher_experience data by counting and sorting them and cutting of the best_n actions from the top.

    Args:
        data: Loaded teacher_experience data.
        env: The environment used to generate the teacher_experience data.
        best_n: The cutoff point determining the maximum number of actions returned.
        plot_choice: Whether the frequency should be plotted.

    Returns:
        The ranked actions as a list.
    """
    decode_action = np.vectorize(lambda act: EncodedTopologyAction.decode_action(act, env))

    # Resort actions
    best_action_freq = data["best_action"].value_counts()
    logging.info(f"There are {len(best_action_freq)} unique actions")

    if plot_choice:
        plt.plot(best_action_freq.values)
        plt.vlines(best_n, ymin=0, ymax=best_action_freq.values.max(), color="red")
        plt.title("Sorted frequency of the actions")
        plt.legend(["Sorted Frequency", "best_n"])

    best_actions = best_action_freq[:best_n]
    best_actions_array = decode_action(best_actions.index.to_numpy())

    return list(best_actions_array)


def rank_actions_tuples(
        data: pd.DataFrame, env: BaseEnv, min_freq_unitary: int = 400, min_freq_tuple: int = 6
) -> List[BaseAction]:
    """Rank the given teacher_experience by sorting them according to their ranking (in the top125 actions list),
    then splitting them into single and tuple actions and mixing both together.

    Args:
        data: Loaded teacher_experience data.
        env: The environment used to generate the teacher_experience of data.
        min_freq_unitary: The minimum frequency of unitary actions to be included in the actionspace.
        min_freq_tuple: The minimum frequency of tuple actions to be included in the actionspace.

    Returns:
        The generated actions as a list.
        The first actions are the tuple actions whereas the later ones are unitary actions.

    """
    decode_action = np.vectorize(lambda act: EncodedTopologyAction.decode_action(act, env))

    # Resort actions based on their ranking and frequency
    top_actions = data.iloc[:, 4:]
    actions = {}

    logging.info("Sorting actions...")
    for rank in range(125):
        top_i = top_actions.iloc[:, rank]
        actions[rank] = top_i

    all_actions = []
    for rank, actions in actions.items():
        for action in actions:
            all_actions.append((rank, action))
    all_actions_df = pd.DataFrame(all_actions, columns=["rank", "action"])

    logging.info("Remove duplicate actions...")
    best_action_freq = all_actions_df["action"].value_counts()

    best_actions = best_action_freq[best_action_freq >= 1]
    best_actions_array = decode_action(best_actions.index.to_numpy())

    best_actions_df = pd.DataFrame({"action": best_actions_array, "count": best_actions})

    logging.info("Counting substations...")
    np_count_actions = np.vectorize(lambda act: len(affected_substations(act)))
    best_actions_df["subs_affected"] = np_count_actions(best_actions_df["action"])

    unique, counts = np.unique(best_actions_df["subs_affected"], return_counts=True)
    logging.info(f"Number of actions: {unique}, counts: {counts}")

    all_unitary_actions = best_actions_df[best_actions_df["subs_affected"] == 1]
    all_tuple_actions = best_actions_df[best_actions_df["subs_affected"] == 2]

    logging.info(f"Filtering by frequency threshold {min_freq_unitary}(unitary) and {min_freq_tuple}(tuple)")
    # Mix and match
    best_unitary_actions = all_unitary_actions[all_unitary_actions["count"] >= min_freq_unitary]
    best_tuple_actions = all_tuple_actions[all_tuple_actions["count"] >= min_freq_tuple]
    mixed_actions = pd.concat([best_tuple_actions, best_unitary_actions])
    logging.info(
        f"actions: {len(mixed_actions)} = {len(best_tuple_actions)}(tuple) + {len(best_unitary_actions)}(unitary)"
    )

    return list(mixed_actions["action"])


def save_actionspace_binbinchen(save_path: Path, actions: List[BaseAction]):
    """Save the given actionspace as a numpy array file (which was used by the binbinchen tutor).

    Args:
        save_path: The path where to save the file to.
        actions: The actionspace to save.

    Returns:
        None.
    """
    sample = actions[0].to_vect()
    action_space = np.zeros((len(actions), sample.shape[0]))
    for i, action in enumerate(actions):
        action_space[i] = action.to_vect()

    np.save(str(save_path), action_space)
    logging.info(f"Save an action space with the size of {action_space.shape[0]:d}")


def make_unitary_actionspace(
        action_space_file_path: Path,
        experience_csv_files: List[Path],
        env_name_path: Union[Path, str],
        best_n: int,
        plot_choice: Optional[bool] = True,
):
    """Using the provided teacher_experience files to generate a unitary space file for the tutor.

    Args:
        action_space_file_path: Where to save the action space file containing the best actions.
        experience_csv_files: The csv files containing the teacher_experience from the teachers.
        env_name_path: The name or path to the environment that is used for teacher_experience generation.
        best_n: The best_n actions to keep.
        plot_choice: Whether the selection process should be plotted for the visualization.

    Returns:
        None, but data is saved under action_space_file_path.
    """

    data = read_experience(experience_csv_files)
    logging.info(f"Read {len(data)} teacher_experience samples from {experience_csv_files}")
    data = filter_good_experience(data)
    logging.info(f"Kept {len(data)} after filtering")

    env = grid2op.make(env_name_path)

    actions = rank_actions_simple(data, env, best_n=best_n, plot_choice=plot_choice)
    save_actionspace_binbinchen(action_space_file_path, actions)
    logging.info(f"Saved {len(actions)} actions after ranking with best_n={best_n} to {action_space_file_path}")


def make_tuple_actionspace(
        action_space_file_path: Path,
        experience_csv_files: List[Path],
        env_name_path: Union[Path, str],
        min_freq_unitary: int,
        min_freq_tuple: int,
):
    """Generate a tuple action space for the tutor using the provided teacher_experience files.

    Args:
        action_space_file_path: Where to save the action space file containing the best actions.
        experience_csv_files: The csv files containing the teacher_experience from the teachers.
        env_name_path: The name or path to the environment that is used for teacher_experience generation.
        min_freq_unitary: The minimum frequency of unitary actions to be included in the actionspace.
        min_freq_tuple: The minimum frequency of tuple actions to be included in the actionspace.

    Returns:
        None, but data is saved under action_space_file_path.
    """

    data = read_experience(experience_csv_files)
    data = filter_good_experience(data)

    env = grid2op.make(env_name_path)

    actions = rank_actions_tuples(data, env, min_freq_unitary=min_freq_unitary, min_freq_tuple=min_freq_tuple)
    save_actionspace_binbinchen(action_space_file_path, actions)
    logging.info(f"Saved {len(actions)} actions to {action_space_file_path}")


def make_seq_actionspace(
        action_space_file_path: Path,
        experience_csv_files: List[Path],
        env_name_path: Union[Path, str],
        seq_length: int,
        best_n: int,
        plot_choice: Optional[bool] = True,
):
    """Generate a tuple action space for the tutor using the provided teacher_experience files.

    Args:
        action_space_file_path: Where to save the action space file containing the best actions.
        experience_csv_files: The csv files containing the teacher_experience from the teachers.
        env_name_path: The name or path to the environment that is used for teacher_experience generation.
        seq_length: Length of the sequential actions that should be kept. E.g. if seq_length=3 than only triple actions.
        are selected.
        best_n: The best_n actions to keep.
        plot_choice: Whether the selection process should be plotted for the visualization.


    Returns:
        None, but data is saved under action_space_file_path.
    """

    data = read_experience(experience_csv_files)
    data = filter_good_experience(data)

    env = grid2op.make(env_name_path)

    actions = rank_actions_seq(data, env, seq=seq_length, best_n=best_n, plot_choice=plot_choice)
    save_actionspace_binbinchen(action_space_file_path, actions)
    logging.info(f"Saved {len(actions)} actions to {action_space_file_path}")


def rank_actions_seq(data: pd.DataFrame, env: BaseEnv, seq: int = 3, best_n: int = 200, plot_choice=False):
    """Rank the given teacher_experience data by counting and sorting them and cutting of the best_n actions from the top.
    This method is used for sequential actions and only selects a specific length of actions.

    Args:
        data: Loaded teacher_experience data.
        env: The environment used to generate the teacher_experience of data.
        seq: Number of actions that has to be within the set.
        best_n: The cutoff point determining the maximum number of actions returned.
        plot_choice: Whether the frequency should be plotted.

    Returns:
        The ranked actions as a list.
    """
    decode_action = np.vectorize(lambda act: EncodedTopologyAction.decode_action(act, env))

    # Resort actions
    best_action_freq = data["best_action"].value_counts()
    logging.info(f"There are {len(best_action_freq)} unique actions")

    if plot_choice:
        plt.plot(best_action_freq.values)
        plt.vlines(best_n, ymin=0, ymax=best_action_freq.values.max(), color="red")
        plt.title("Sorted frequency of the actions")
        plt.legend(["Sorted Frequency", "best_n"])

    best_actions_array = list(decode_action(best_action_freq.index.to_numpy()))
    best_actions_array = [a for a in best_actions_array if len(a.as_dict()["set_bus_vect"]["modif_subs_id"]) == seq]
    best_actions_array = best_actions_array[:best_n]

    return best_actions_array




"""In this file, a neural network is developed to fit the dataset generated by Tutor.
Depending on the observation space and action space, the tutor model can/has to be
adjusted.

The Junior model returns a one-hot encoded output, based on the number of actions.

Credit: The junior is a more general approach of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""
import logging
from pathlib import Path
from typing import Union, Optional, Tuple, List

import grid2op.Environment
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler

from curriculumagent.junior.junior_student_pytorch import JuniorParamTorch, JuniorTorch
from curriculumagent.junior.junior_student_tensorflow import JuniorParamTF, JuniorTF


class Junior:
    """
    Create the overall Junior which can train either a torch or a tensorflow model.
    """

    def __init__(self,
                 action_space_file: Union[Path, List[Path], str],
                 run_with_tf: bool = True,
                 config: Union[JuniorParamTF, JuniorParamTorch] = {},
                 seed: Optional[int] = None,
                 scaler: Optional[StandardScaler] = None, #ToDo ask about scaler
                 run_nni: bool = False
                 ):
        """Constructor of the Junior with either Pytorch Lightning or with Keras-Tensorflow.
        The junior can receive hyperparameters in the config to create a more flexible model.

        Furhter, it is possible to create a hyperparameter search with tune ore with NNI.

        Except setting all variables, the init additionally requires the size of the train set and optionally
        the number of epochs.

        Note:
            Pass epochs, learning_rate, batchsize, layer_size in the config.

        Args:
            run_with_tf: Should the agent run with tensorflow. If set to False, torch is selected instead.
            action_space_file: Action Space file that was used for the Tutor training. This is needed to extract the
            correct shape of the Junior model.
            config: Dictionary containing the correct input for the hyperparameters.
            seed: Optional Seed to reproduce results.
            run_nni: Whether NNI is used. If True, then a specific callback is added.

        Returns:
            None.
        """

        # Set self actions to a list to iterate for later
        if run_with_tf:
            self.junior_model = JuniorTF(action_space_file=action_space_file,
                                         config=config,
                                         seed=seed,
                                         run_nni=run_nni)
        else:
            self.junior_model = JuniorTorch(action_space_file=action_space_file,
                                            config=config,
                                            seed=seed,
                                            run_nni=run_nni)
        self.run_with_tf = run_with_tf
        self.actions = None
        if isinstance(action_space_file, Path):
            assert action_space_file.is_file()
            self.actions = np.load(str(Path(action_space_file)))

        elif isinstance(action_space_file, list):
            for act_path in action_space_file:
                assert act_path.is_file()
            self.actions = np.concatenate([np.load(str(act_path)) for act_path in action_space_file], axis=0)

        self.scaler = scaler if isinstance(scaler, BaseEstimator) else None

    def train(self,
              run_name: str,
              dataset_path: Path,
              target_model_path: Path,
              dataset_name: str = "junior_dataset",
              epochs: int = 1000,
              **kwargs
              ) -> dict:
        """Loads the dataset and then trains the JuniorModel with either Tensorflow or torch.


        Args:
            run_name: The name of the training run.
            dataset_path: Path to the dataset files.
            target_model_path: Path, where to save the model.
            action_space_file: Optional action space file of the tutor.
            dataset_name: The name of the dataset in {dataset_name}_train.npz.
            epochs: The number of epochs to train.
            kwargs: Additional arguments for either the Torch or TF junior

        Returns:
            Training history with at least the keys accuracy and val_accuracy

        """
        if not target_model_path.is_dir():
            logging.warning(f"{target_model_path} does not exists yet. Create directory")
            target_model_path.mkdir(parents=True, exist_ok=True)

        ckpt_dir = target_model_path / f"ckpt-{run_name}"

        if self.run_with_tf:
            s_train, a_train, s_validate, a_validate, s_test, a_test = load_dataset(dataset_path, dataset_name)

            if self.scaler:
                s_train = self.scaler.transform(s_train)
                s_validate = self.scaler.transform(s_validate)
            # Get maximum number of actions:

            history = self.junior_model.train(
                x_train=s_train,
                y_train=a_train,
                x_validate=s_validate,
                y_validate=a_validate,
                log_dir=None,
                ckpt_dir=ckpt_dir,
                epochs=epochs,
                **kwargs
            )
            # Save model
            self.junior_model.model.save(target_model_path)

        else:
            history = self.junior_model.train(
                dataset_path=dataset_path,
                target_model_path=ckpt_dir,
                dataset_name=dataset_name,
                epochs=epochs,
                scaler=self.scaler,
                **kwargs)

        return history

    def test(self,
             dataset_path: Path,
             checkpoint_path: Optional[Path] = None,
             dataset_name: str = "junior_dataset"
             ) -> dict:
        """Test the given checkpoint against the test data set.

        Args:
            checkpoint_path: The checkpoint file to use for the conversion/input.
            dataset_path: Path to the dataset used to train the checkpoint.
            dataset_name: The name of the dataset in {dataset_name}_test.npz.

        Returns:
            Dictionary with accuracy values achieved on the testing dataset.

        """
        logging.info(f"A total of {len(self.actions)} are assumed, based on the action_space_file input.")

        if self.run_with_tf:
            _, _, _, _, s_test, a_test = load_dataset(dataset_path, dataset_name)

            if self.scaler:
                s_test = self.scaler.transform(s_test)

            accuracy = self.junior_model.test(x=s_test, y=a_test, save_path=checkpoint_path)

        else:

            if self.scaler:
                accuracy = self.junior_model.test(dataset_path=dataset_path, dataset_name=dataset_name, scaler=self.scaler, checkpoint_path=checkpoint_path)
            else:
                accuracy = self.junior_model.test(dataset_path=dataset_path, dataset_name=dataset_name, checkpoint_path=checkpoint_path)
        return accuracy


def load_dataset(
        dataset_path: Union[str, Path], dataset_name: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Function for loading the datasets of the Tutor run.

    Note:
        All datasets (training, validation and testing) have the same name but different endings.
        As example:
        dataset_name = data
        training: data_train.npz
        validation: data_val.npz
        test: data_test.npz

    Args:
        dataset_path: Path where to find the Tutor results.
        dataset_name: Name of the tutor results.

    Returns:
        Tuple, containing vectors (numpy arrays) with training, val and test data.

    """
    train_path = Path(dataset_path) / f"{dataset_name}_train.npz"
    train_data = np.load(train_path)

    validation_path = Path(dataset_path) / f"{dataset_name}_val.npz"
    validation_data = np.load(validation_path)

    test_path = Path(dataset_path) / f"{dataset_name}_test.npz"
    test_data = np.load(test_path)

    return (
        train_data["s_train"],
        train_data["a_train"],
        validation_data["s_validate"],
        validation_data["a_validate"],
        test_data["s_test"],
        test_data["a_test"],
    )



"""In this file all utility methods are collected which are jointly used by the teacher, tutor, junior and senior agent.
The major task of the methods is to communicate with the Grid2Op Environment.

Credit: The methods are the enhanced methods of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""

import logging
from copy import deepcopy
from typing import Optional, List, Tuple, Iterator

import grid2op
import numpy as np
from grid2op.Action import BaseAction, ActionSpace
from grid2op.Environment import BaseEnv
from grid2op.Observation import BaseObservation
from grid2op.dtypes import dt_int


def find_best_line_to_reconnect(
        obs: BaseObservation, original_action: BaseAction) -> BaseAction:
    """Given an observation and action try to reconnect a line by modifying the given original action
    and returning the modified action with the reconnection if possible.

    Args:
        obs: The current observation of the agent.
        original_action: The action the agent is going to take.

    Returns:
        The modified original_action which tries to reconnect disconnected lines.

    """
    disconnected_lines = np.where(obs.line_status == False)[0]
    if len(disconnected_lines) == 0:
        return original_action
    min_rho = 10
    line_to_reconnect = -1
    for line in disconnected_lines:
        if not obs.time_before_cooldown_line[line]:
            reconnect_array = np.zeros_like(obs.rho, dtype=int)
            reconnect_array[line] = 1
            reconnect_action = deepcopy(original_action)
            reconnect_action.update({"set_line_status": reconnect_array})
            if not is_legal(reconnect_action, obs):
                continue
            o, _, done, info = obs.simulate(reconnect_action)
            if not is_valid(observation=obs, act=reconnect_action, done_sim=done, info_sim=info):
                continue
            if o.rho.max() < min_rho:
                line_to_reconnect = line
                min_rho = o.rho.max()

    reconnect_out = deepcopy(original_action)
    if line_to_reconnect != -1:
        reconnect_array = np.zeros_like(obs.rho, dtype=int)
        reconnect_array[line_to_reconnect] = 1
        reconnect_out.update({"set_line_status": reconnect_array})

    return reconnect_out


def is_legal(action: BaseAction, obs: BaseObservation) -> bool:
    """Return true if the given action is valid under the current given observation.

    Args:
        action: The action to check for.
        obs: The current observation of the environment.

    Returns:
        Whether the given action is valid/legal or not.

    """
    action_dict = action.as_dict()

    if action_dict == {}:
        return True

    topo_action_type = list(action_dict.keys())[0]
    legal_act = True

    # Check substations:
    if topo_action_type == "set_bus_vect" or topo_action_type == "change_bus_vect":
        substations = [int(sub) for sub in action.as_dict()[topo_action_type]["modif_subs_id"]]

        for substation_to_operate in substations:
            if obs.time_before_cooldown_sub[substation_to_operate]:
                # substation is cooling down
                legal_act = False
            # Check lines:
            for line in [
                val["id"]
                for key, val in action.as_dict()[topo_action_type][str(substation_to_operate)].items()
                if "line" in val["type"]
            ]:
                if obs.time_before_cooldown_line[line] or not obs.line_status[line]:
                    # line is cooling down, or line is disconnected
                    legal_act = False
    elif topo_action_type == "set_line_status":

        lines = [int(line) for line in action.as_dict()[topo_action_type]["connected_id"]]
        for line in lines:
            if obs.time_before_cooldown_line[line]:
                legal_act = False

    return legal_act


def check_convergence(action: BaseAction, obs: BaseObservation) -> bool:
    """Return true if the given action leads to convergence in the network under the current given observation.

    Args:
        action: The action to check for.
        obs: The current observation of the environment.

    Returns:
        Whether the given action leads to convergent state in the network or not.

    """

    simulator = obs.get_simulator()

    load_p_stressed = obs.load_p * 1.05
    gen_p_stressed = obs.gen_p * 1.05

    simulator_stressed = simulator.predict(act=action, new_gen_p=gen_p_stressed, new_load_p=load_p_stressed)

    if not simulator_stressed.converged:
        return False
    else:
        return True


def extract_action_set_from_actions(
        action_space: ActionSpace, action_vect: np.ndarray
) -> List[BaseAction]:
    """Method to separate multiple substation actions into single actions.

    We use the decompose_as_unary_actions() method from grid2op.

    Args:
        action_space: action space of Grid2Op environment.
        action_vect: single action from numpy array.

    Returns:
        List with actions.

    """
    action_set = [action_space({})]


    act: grid2op.Action.BaseAction = action_space.from_vect(action_vect)

    act_dec = act.decompose_as_unary_actions()

    if act_dec:
        action_set =[]

        for v in act_dec.values():
            action_set += v

    return action_set



def split_action_and_return(
        obs: BaseObservation, action_space: ActionSpace, action_vect: np.ndarray
) -> Iterator[BaseAction]:
    """Split an action with potentially multiple affected stations and return them sequentially as a generator/iterator.

    Depending on the input, the method either executes the numpy array as a unitary step, or
    if the input requires multiple steps (i.e. in case of a tuple or triple action) the method
    orders the actions and selects the best choice, then executes the actions sequentially.

    Note that if multiple steps are computed only the last observation is return. Further, the
    reward then consists of the cumulative reward. As Example: The action is a triple action. Thus, the
    reward is the cumulative reward of all three steps.

    All actions are checked for line reconnect.

    Args:
        action_space: action space of Grid2Op environment.
        obs: Current Observation.
        action_vect: Teacher Action that can either be a unitary, tuple or triple action.

    Returns:
        The next best action to execute, without the line connect!!!

    """

    # Special case: Do nothing
    d_n = action_space({})

    if not action_vect.any() or action_space.from_vect(action_vect) == d_n:
        yield d_n
        return # We must call return or else we call it again

    # First extract action:
    split_actions = extract_action_set_from_actions(action_space, action_vect)
    # Now simulate through all actions:
    for _ in range(len(split_actions)):
        # Iterate through remaining actions and choose the best one to execute next
        obs_min = np.inf
        best_choice = None

        for act in split_actions:
            act_plus_reconnect = find_best_line_to_reconnect(obs, act)
            obs_f, _, done, info = obs.simulate(act_plus_reconnect)

            # Check if valid
            if not is_valid(observation=obs, act=act_plus_reconnect, done_sim=done, info_sim=info):
                continue

            if obs_f.rho.max() < obs_min:
                best_choice = act
                obs_min = obs_f.rho.max()

        if best_choice is None:
            best_choice = d_n

        # Assert whether a reconnection of the lines might be
        yield best_choice
        if best_choice in split_actions:
            split_actions.remove(best_choice)


def is_valid(observation: BaseObservation, act: BaseAction, done_sim, info_sim, check_overload: Optional[bool] =
False) -> bool:
    """Checks whether the simulation output is legal/valid for our configuration

    Args:
        observation: original observation
        act: action of the agent
        done_sim: done variable of the obs.simulate()
        info_sim: info variable of the obs.simulate()
        check_overload: Boolean, whether to simulate a stress of the generation and load

    Returns:
        bool: Whether the action is valid.

    """
    valid_action = True
    if done_sim:
        valid_action = False
    if not is_legal(act, observation):
        valid_action = False
    if info_sim["is_illegal"]:
        valid_action = False
    if info_sim['is_ambiguous']:
        valid_action = False
    if any(info_sim["exception"]):
        valid_action = False
        # first check convergence of the stressed net after action
    if check_overload and not check_convergence(act, observation):
        valid_action = False
    return valid_action


def simulate_action(action_space: ActionSpace, obs: BaseObservation, action_vect: np.array,
                    check_overload: Optional[bool] = False) -> \
        Tuple[
            float, bool]:
    """Simulate the impact of the given action. This method can be used by both  the tutor and the
    Senior and should result in better choices, when tuple/tripple actions are implemented.

    Args:
        action_space: Action Space of the environment.
        obs: Current Observation of the Grid2Op environment.
        action_vect: Array that can be converted back to Grid2Op Actions. Tuples and Tripples are supported.
        check_overload: Boolean, whether to simulate a stress of the generation and load

    Returns:
        The max Obs of the action.

    """
    action = action_space.from_vect(action_vect)
    act_dict = action.as_dict()
    if "set_bus_vect" not in act_dict.keys():
        action = find_best_line_to_reconnect(obs=obs, original_action=action)

        obs_f, _, done, info = obs.simulate(action)
    elif len(act_dict["set_bus_vect"]["modif_subs_id"]) == 1:
        action = find_best_line_to_reconnect(obs=obs, original_action=action)
        obs_f, _, done, info = obs.simulate(action)
    else:
        # length is longer than 1:
        gen = split_action_and_return(obs=obs, action_space=action_space, action_vect=action_vect)
        action = next(gen)
        obs_f, _, done, info = obs.simulate(action)
    rho_max = obs_f.rho.max()

    valid_action = is_valid(observation=obs, act=action, done_sim=done, info_sim=info, check_overload=check_overload)

    return rho_max, valid_action


def split_and_execute_action(env: BaseEnv, action_vect: np.ndarray) -> Tuple[BaseObservation, float, bool, dict]:
    """Split and execute an action with potentially multiple affected stations.

    Depending on the input, the method either executes the numpy array as a unitary step, or
    if the input requires multiple steps (i.e. in case of a tuple or tripple action) the method
    does multiple steps.

    Note that if multiple steps are computed only the last observation is return. Further, the
    reward then consists of the cummulative reward. As Example: The action is a tripple action. Thus, the
    reward is the cummulative reward of all three steps.

    All actions are checked for line reconnect.

    Args:
        env: Grid2Op Environment.
        action_vect: Teacher Action that can either be a unitary, tuple or tripple action.

    Returns: Output of the env.step() function consisting of an Observation, (cumulative) Reward, a Done statement
    and info.

    """

    # First extract action:
    split_actions = extract_action_set_from_actions(env.action_space, action_vect)
    # Now simulate through all actions:
    cum_rew = 0
    done = env.done
    obs = env.get_obs()
    info = {}
    for _ in range(len(split_actions)):
        # Iterate through remaining actions and choose the best one to execute next
        obs_min = np.inf
        best_choice = find_best_line_to_reconnect(obs, env.action_space({}))

        chosen_act = None
        for act in split_actions:
            act_plus_reconnect = find_best_line_to_reconnect(obs, act)
            if not is_legal(act_plus_reconnect, obs):
                continue

            obs_, _, done_sim, _ = obs.simulate(act_plus_reconnect)
            if obs_.rho.max() < obs_min and done_sim is False:
                best_choice = act_plus_reconnect
                chosen_act = act
                obs_min = obs_.rho.max()


        # Assert whether a reconnection of the lines might be

        obs, reward, done, info = env.step(best_choice)
        cum_rew += reward
        if done:
            break
        if chosen_act:
            split_actions.remove(chosen_act)

    return obs, cum_rew, done, info


def revert_topo(action_space: grid2op.Action.ActionSpace, obs: grid2op.Observation.BaseObservation,
                rho_limit: Optional[float] = 0.8) -> np.array:
    """Method, if the topology can be reverted to its original state.

    Given that the original state is all substations at bus bar 1, we search for any bus bars equal to
    2. If one is found, we check whether a reversion of the substation reduces the rho.max(). To
    ensure that the minimization only happens, when the grid is stable, we propose a
    rho_limit. If the rho.max() is above the limit, no reversion takes place!

    Args:
        action_space: Action Space of the environment.
        obs: Current Observation of the Grid2Op Environment.
        rho_limit: Optional Limit for the topology search.

    Returns:
        Array of topo action.

    """
    act_best, obs_sim, idx = None, None, None
    if np.any(obs.topo_vect == 2) and np.all(obs.line_status):
        available_actions = {}
        for sub_id in range(obs.n_sub):
            if np.any(obs.sub_topology(sub_id) == 2):
                available_actions[sub_id] = action_space(
                    {"set_bus": {"substations_id": [(sub_id, np.ones(len(obs.sub_topology(sub_id)), dtype=dt_int))]}}
                ).to_vect()

        if any(available_actions):
            min_rho = min([obs.rho.max(), rho_limit])

            for sub_id, act_a in available_actions.items():
                obs_sim, valid_action = simulate_action(action_vect=act_a, action_space=action_space, obs=obs)
                if not valid_action:
                    continue
                if obs_sim < min_rho:
                    idx = sub_id
                    act_best = act_a.copy()
                    rho_best = obs_sim.copy()

    if act_best is not None:
        logging.info(
            f"Revert Substation {idx} from {available_actions.keys()} to original with current rho "
            f" {obs.rho.max()} and {rho_best} as new. Line status:{np.all(obs.line_status)}")
        return act_best
    else:
        return action_space({}).to_vect()





def map_actions(list_of_actions: List[np.ndarray]) -> List[dict]:
    """This method iterates over the list of actions and passes an id for all actions. This helps for
    the junior and the senior evaluation.

    Args:
        list_of_actions: List that contains the actions in numpy format.

    Returns:
        List that contains a dictionary with the id as key and the action as value.

    """
    total_action = []
    idx = 0
    for actions in list_of_actions:
        act_id = {}
        for i, act in enumerate(actions):
            act_id[i + idx] = act
        idx += len(act_id)
        total_action.append(act_id)

    return total_action


def change_bus_from_topo_vect(tv1, tv2, action_space):
    """
    Get the topology action in form of change_bus in order to get from topo_vect tv1 to topo_vect tv2.
    Note that we do not (!) change the -1 values, given that this would make the action illegal

    Args:
        tv1: original topology
        tv2: topology we want to move to
        env: grid2op environment

    Returns: Returns a action that can also effect multiple substations! Thus, it is essential to split
    them into unitary actions

    """
    # We only look at changes that are not(!) -1
    ids = np.where(np.abs(tv1 - tv2) == 1)[0]
    a = action_space({})
    a.change_bus = list(ids)
    return a


def set_bus_from_topo_vect(tv1, tv2, action_space):
    """
    Get the topology action in form of set_bus in order to get from topo_vect tv1 to topo_vect tv2.
    Note that we do not (!) change the -1 values, given that this would make the action illegal

    Args:
        tv1: original topology
        tv2: topology we want to move to
        env: grid2op environment

    Returns: Returns a action that can also effect multiple substations! Thus, it is essential to split
    them into unitary actions

    """
    # We only look at changes that are not(!) -1
    #ToDo: What about 2-1 and 1-2
    ids = np.where(np.abs(tv1 - tv2) == 1)[0]

    set_bus = [(i, tv2[i]) for i in ids]

    a = action_space({})
    a.set_bus = set_bus
    return a


"""Script to score and evaluate the final agent and compare it to the original scores.
"""
import json
import logging
import os
import pickle
import shutil
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import grid2op
import lightsim2grid
import numpy as np
import pandas as pd
import tensorflow as tf
from grid2op.Agent import BaseAgent
from grid2op.Environment import BaseEnv
from grid2op.Episode import EpisodeData
from grid2op.Reward import L2RPNWCCI2022ScoreFun, L2RPNSandBoxScore
from grid2op.dtypes import dt_int
from grid2op.utils import ScoreL2RPN2020, EpisodeStatistics
from ubelt import Timer

env_name = "l2rpn_neurips_2020_track1"
test_flag = False


@dataclass
class AgentReport:
    """Report that consists of data gathered from evaluating an agent, mainly used for comparing agents and
    tracking performance.
    """

    agent_name: str
    score_data: Dict
    nb_episodes: int
    avg_score: Optional[float] = None
    evaluation_time: Optional[float] = None
    g2op_version: Optional[str] = None

    @staticmethod
    def load(path: Path) -> "AgentReport":
        """Load a report previously saved using :meth:`save`.

        Args:
            path: Path to load the report file from.

        Returns:
            Report.
        """

        with path.open("rb") as f:
            report = pickle.load(f)
        return report

    def save(self, path: Path):
        """Save the report into a machine-readable format for reloading with :meth:`load`.

        Args:
            path: Where to save the report file.

        Returns:
            None.
        """
        with path.open("wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)


def render_report(report_path: Path, report: AgentReport, comparison_reports: List[AgentReport]):
    """Render the report as a markdown file for easy viewing by humans.

    Args:
        report_path: The path to save the report to.

    Returns:
        None.
    """
    with report_path.open("w") as f:
        f.write(f"# {report.agent_name} Performance Report\n\n")
        f.write(f"**{report.agent_name} Score: {report.avg_score}**\n\n")

        for comparison_report in comparison_reports:
            f.write(f"*{comparison_report.agent_name} Score: {comparison_report.avg_score}*\n\n")

        f.write("\n\n ## Plots\n\n")

        survival_plot = plot_bar_comparison([report] + comparison_reports, "ts_survived")
        score_plot = plot_bar_comparison([report] + comparison_reports, "all_scores")

        survival_plot_name = f"{report.agent_name}_survival.png"
        score_plot_name = f"{report.agent_name}_score.png"
        survival_plot.figure.savefig(report_path.parent / survival_plot_name, bbox_inches="tight")
        score_plot.figure.savefig(report_path.parent / score_plot_name, bbox_inches="tight")

        f.write(f"\n![Score Plot]({score_plot_name})\n")
        f.write(f"\n![Survival Plot]({survival_plot_name})\n")

        f.write("\n\n ## Raw Data/Scores \n\n")
        pd.DataFrame(report.score_data).to_markdown(f)

        f.write(f"\n\n ## Metadata: \n\n Grid2Op Version: {report.g2op_version}")


def plot_bar_comparison(reports: List[AgentReport], metric_key: str) :
    """Create Bar Plot of the performance.

    Args:
        reports: Reports of the agents.
        metric_key: Metric key, where to evaluate.

    Returns:
        Plot.

    """
    assert len(reports) >= 2, "Need at least 2 reports for comparison"

    score_dfs = [(r.agent_name, pd.DataFrame(r.score_data).set_index(("episode_name"))[metric_key]) for r in reports]
    names, comb_series = zip(*score_dfs)
    score_merged = pd.concat(comb_series, axis=1)
    score_merged.columns = names

    axes = (score_merged.plot.bar(title=f"{metric_key} Comparison", ylabel=metric_key, figsize=(10, 6))
            .legend(loc="center left", bbox_to_anchor=(1, 0.5))

            )
    return axes

class ScoreL2RPN2020WithNames(ScoreL2RPN2020):
    """Extension of ScoreL2RPN2020 class to return scenario names as well.

    """
    def get(
            self, agent: BaseAgent, path_save: Optional[str] = None, nb_process: int = 1,
            seed_name:str = None) -> Tuple[List[float], List[int], List[int], List[str]]:
        """Get the score of the agent depending on what has been computed.

        Args:
            agent: The agent you want to score.
            path_save: The path where you want to store the logs of your agent.
            nb_process: Number of process to use for the evaluation.

        Returns:
            all_scores: List of your agent per scenarios score.
            ts_survived: List of the step number your agent successfully managed for each scenario.
            total_ts: Total step number for each scenario.
            scenario_names: Names for each scenario executed.
            seed_name: When running multiple seeds, it is possilbe to specify the name in order to save
            multiple do_nothing reports, depending on the seed of the env.
        """
        if path_save is not None:
            need_delete = False
            path_save = os.path.abspath(path_save)
        else:
            need_delete = True
            dir_tmp = tempfile.TemporaryDirectory()
            path_save = dir_tmp.name

        if self.verbose >= 1:
            print("Starts the evaluation of the agent")

        es_instance = EpisodeStatistics(self.env, name_stats=seed_name)

        es_instance.run_env(
            self.env,
            env_seeds=self.env_seeds,
            agent_seeds=self.agent_seeds,
            path_save=path_save,
            parameters=self.env.parameters,
            scores_func=self.scores_func,
            agent=agent,
            max_step=self.max_step,
            nb_scenario=self.nb_scenario,
            pbar=self.verbose >= 2,
            nb_process=nb_process,
        )
        if self.verbose >= 1:
            print("Start the evaluation of the scores")

        meta_data_dn = self.stat_dn.get_metadata()
        no_ov_metadata = self.stat_no_overflow_rp.get_metadata()

        all_scores = []
        ts_survived = []
        total_ts = []
        scenario_names = []
        for ep_id in range(self.nb_scenario):
            this_ep_nm = meta_data_dn[f"{ep_id}"]["scenario_name"]
            with open(os.path.join(path_save, this_ep_nm, EpisodeData.META), "r", encoding="utf-8") as f:
                this_epi_meta = json.load(f)
            with open(os.path.join(path_save, this_ep_nm, EpisodeData.OTHER_REWARDS), "r", encoding="utf-8") as f:
                this_epi_scores = json.load(f)
            score_this_ep, nb_ts_survived, total_ts_tmp = self._compute_episode_score(
                ep_id,
                meta=this_epi_meta,
                other_rewards=this_epi_scores,
                dn_metadata=meta_data_dn,
                no_ov_metadata=no_ov_metadata,
            )
            all_scores.append(score_this_ep)
            ts_survived.append(nb_ts_survived)
            total_ts.append(total_ts_tmp)
            scenario_names.append(this_ep_nm)

        if need_delete:
            dir_tmp.cleanup()
        return all_scores, ts_survived, total_ts, scenario_names


class ScoreL2RPN2022WithNames(ScoreL2RPN2020WithNames):
    """Extension of ScoreL2RPN2020WithNames class with the initialisation of the ScoreL2RPN2022 values.

    The only difference is that it takes a different score function. See `ScoreL2RPN2022
    <https://grid2op.readthedocs.io/en/latest/_modules/grid2op/utils/l2rpn_wcci_2022_scores.html#ScoreL2RPN2022>`_
    """

    def __init__(self,
                 env,
                 env_seeds=None,
                 agent_seeds=None,
                 nb_scenario=16,
                 min_losses_ratio=0.8,
                 verbose=0, max_step=-1,
                 nb_process_stats=1,
                 scores_func=L2RPNWCCI2022ScoreFun,
                 score_names=None):
        super().__init__(env, env_seeds, agent_seeds, nb_scenario, min_losses_ratio, verbose, max_step,
                         nb_process_stats, scores_func, score_names)


def score_agent(
        agent: BaseAgent,
        env: BaseEnv,
        log_path: Path,
        seed: int = 42,
        name: Optional[str] = None,
        nb_episodes: Optional[int] = None,
        nb_process: int = os.cpu_count(),
        reinit: Optional[bool] = False,
        score_l2rpn2020: Optional[bool] = True
) -> AgentReport:
    """Score the given agent in the given environment, saving logs in log_dir.

    Args:
        agent: The agent to score.
        env: The environment to score the agent on.
        log_path: The directory to save agent logs in.
        seed: Seed used to score the agent.
        name: The name of the agent.
        nb_episodes: How many episodes/chronics the agent should be scored. Equals 2 by default.
        nb_process: How many processes should be used to perform the scoring.
                    Some agents may support one process only due to unsupported serialization/pickling of state.
        reinit: Optional Boolean, whether the underlying environment statistics should be deleted and
                recalculated.
        score_l2rpn2020: Optional Boolean Indicating, which score should be used for the evaluation. If True the
                score_agent method will use the ScoreL2RPN2020 metric, else ScoreL2RPN2022.

    Returns:
        The :class:`AgentReport` describing the agent performance.

    """
    g2op_version = str(grid2op.__version__)
    ls2g_version = str(lightsim2grid.__version__)
    print(f"Grid2Op version: {g2op_version}")
    print(f"lightsim2grid version: {ls2g_version}")
    try:
        # Disable all GPUS
        tf.config.set_visible_devices([], "GPU")
        visible_devices = tf.config.get_visible_devices()
        for device in visible_devices:
            assert device.device_type != "GPU"
    except Exception as e:
        # Invalid device or cannot modify virtual devices once initialized.
        logging.warning(f"We caught the exception {e}")
        pass

    # Set random Seed
    if not nb_episodes:
        nb_episodes = 2
    np.random.seed(seed)
    max_int = np.iinfo(dt_int).max
    env_seeds = list(np.random.randint(max_int, size=nb_episodes))
    agent_seeds = list(np.random.randint(max_int, size=nb_episodes))

    # Evaluate trained agent
    if name:
        agent_name = f"{name}_{agent.__class__.__name__}"
    else:
        agent_name = f"{agent.__class__.__name__}"
    print(f"Evaluating agent with name {agent_name}")

    # Start scoring the agent and time the execution
    timer = Timer()
    with timer:
        if reinit:
            li_stats = EpisodeStatistics.list_stats(env)
            print(f"Deleting the following environment statistics for seed {seed}:")
            print(li_stats)

            for path, el in li_stats:
                shutil.rmtree(os.path.join(path, el), ignore_errors=True)
            print("Delition done! Start re-running agent")

        if score_l2rpn2020:
            my_score = ScoreL2RPN2020WithNames(
                env, nb_scenario=nb_episodes, env_seeds=env_seeds, agent_seeds=agent_seeds, verbose=3
            )
        else:
            my_score = ScoreL2RPN2022WithNames(
                env, nb_scenario=nb_episodes, env_seeds=env_seeds, agent_seeds=agent_seeds, verbose=3
            )

        log_path.mkdir(exist_ok=True, parents=True)
        all_scores, ts_survived, total_ts, episode_names = my_score.get(
            agent, nb_process=nb_process, path_save=str(log_path),seed_name=str(seed)
        )

    score_data = {
        "all_scores": all_scores,
        "ts_survived": ts_survived,
        "total_ts": total_ts,
        "episode_name": episode_names,
    }
    avg_score = np.array(all_scores).mean()
    return AgentReport(
        agent_name=agent_name,
        score_data=score_data,
        nb_episodes=nb_episodes,
        evaluation_time=timer.elapsed,
        g2op_version=str(grid2op.__version__),
        avg_score=avg_score,
    )


def load_or_run(
        agent: BaseAgent,
        env: BaseEnv,
        output_path: Path,
        name: Optional[str] = None,
        nb_processes: int = os.cpu_count(),
        overwrite: bool = False,
        number_episodes: int = 2,
        seed: int = 42,
        reinit: bool = False,
        score_l2rpn2020: Optional[bool] = True
) -> AgentReport:
    """Load the given report at cache_path or score the agent if it doesn't exist.

    For easier comparison, this function tries to load cached results of the :func:`score_agent` function.
    However, it does not check whether the result is out of date. So if you updated the agent to score, you either have
    to delete the corresponding cache file or set the overwrite parameter to True.

    Args:
        agent: The agent to score.
        env: The environment to score the agent in.
        output_path: The path to a directory where cached results as well as agent logs of the scoring will be stored.
        name: The name of the report. The agent class name will be used if this is not set.
        nb_processes: How many processes should be used to perform the scoring.
            Some agents may only support one process due to unsupported serialization/pickling of state.
        overwrite: If set to true, ignore and overwrite the cached result.
        number_episodes: How many episodes to run for comparison.
        seed: Optional seed.
        reinit: Optional Boolean, whether the underlying environment statistics should be deleted and
                recalculated.
        score_l2rpn2020: Optional Boolean, indicating which score should be used for the evaluation. If True the
                score_agent() method will use the ScoreL2RPN2020 metric, else the ScoreL2RPN2022.

    Returns:
        The :class:`AgentReport` describing the agent performance.

    """
    if not name:
        name = agent.__class__.__name__
    output_path.mkdir(exist_ok=True, parents=True)
    report_path = output_path / Path(f"{name}_report_data.pkl")
    if report_path.exists() and not overwrite and not reinit:
        logging.info(f"Using cached results from {report_path}")
        report = AgentReport.load(report_path)
        return report

    logs_path = output_path / "agent_logs" / f"{name}"

    report = score_agent(
        agent, env, log_path=logs_path, name=name, nb_process=nb_processes, nb_episodes=number_episodes,
        seed=seed, reinit=reinit, score_l2rpn2020=score_l2rpn2020
    )
    report.save(report_path)
    return report



"""This file consist of the two tutor approaches. The general_tutor method is the more general approach.
that enables multiple tweaks in order to enhance the training.
The original_tutor method is the original approach by @https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution

"""

from pathlib import Path
from typing import Union, Optional, List

from curriculumagent.tutor.collect_tutor_experience import generate_tutor_experience
from curriculumagent.tutor.tutors.n_minus_one_tutor import NminusOneTutor


def general_tutor(
        env_name_path: Union[Path, str],
        save_path: Union[Path, str],
        action_paths: Union[Path, List[Path]],
        num_chronics: Optional[int] = None,
        num_sample: Optional[int] = None,
        jobs: int = -1,
        seed: Optional[int] = None,
):
    """Method to run the general Tutor in parallel

    Args:
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        save_path: Where to save the teacher_experience.
        action_paths: List of action sets (in .npy format).
        num_chronics: Total numer of chronics.
        num_sample: Length of sample from the num_chronics. With replacement!
        jobs: Number of jobs in parallel.
        seed: Whether to set a seed to the sampling of environments.

    Returns:
        None, saves results as numpy file.

    """
    generate_tutor_experience(
        env_name_path=env_name_path,
        save_path=save_path,
        action_paths=action_paths,
        num_chronics=num_chronics,
        num_sample=num_sample,
        jobs=jobs,
        seed=seed,
    )


def n_minus_1_tutor(
        env_name_path: Union[Path, str],
        save_path: Union[Path, str],
        action_paths: Union[Path, List[Path]],
        num_chronics: Optional[int] = None,
        num_sample: Optional[int] = None,
        jobs: int = -1,
        seed: Optional[int] = None,
        revert_to_original_topo: Optional[bool] = False,
):
    """Method to run the general Tutor in parallel.

    Args:
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        save_path: Where to save the teacher_experience.
        action_paths: List of action sets (in .npy format).
        num_chronics: Total numer of chronics.
        num_sample: Length of sample from the num_chronics. With replacement!
        jobs: Number of jobs in parallel.
        seed: Whether to set a seed to the sampling of environments.
        revert_to_original_topo: Whether the N-1 Tutor should revert to original topology
        within the optimization process.
.
    Returns:
        None, saves results as numpy file.

    """

    if revert_to_original_topo:
        tutor_kwargs = {"revert_to_original_topo": True}
    else:
        tutor_kwargs = None

    generate_tutor_experience(
        env_name_path=env_name_path,
        save_path=save_path,
        action_paths=action_paths,
        num_chronics=num_chronics,
        num_sample=num_sample,
        jobs=jobs,
        seed=seed,
        TutorAgent=NminusOneTutor,
        tutor_kwargs=tutor_kwargs,
    )



import json
import logging
import os
import pickle
import random
from pathlib import Path
from typing import Union, List, Optional

import ray
import tensorflow as tf
from ray._raylet import ObjectRef
from ray.rllib.algorithms.ppo import PPOConfig
from ray.rllib.models import ModelCatalog
from ray.rllib.utils import check_env
from sklearn.base import BaseEstimator

from curriculumagent.senior.rllib_execution.senior_env_rllib import SeniorEnvRllib
from curriculumagent.senior.rllib_execution.senior_model_rllib import Grid2OpCustomModelTF, Grid2OpCustomModelTorch
from curriculumagent.submission.my_agent import MyAgent
import torch

class Senior:
    """
    This class is the Senior agent. It is based on the open source framework RLlib and trains with the
    PPO. The Senior model requires a Junior model and ray to be initialized.

    Note that you can use this class a parent, if you want to change the default values of the
    underlying rllib_environment or the model

    """

    def __init__(self,
                 env_path: Union[str, Path],
                 action_space_path: Union[Path, List[Path]],
                 model_path: Union[Path, str],
                 ckpt_save_path: Optional[Union[Path, str]] = None,
                 run_with_tf: bool = True,
                 scaler: Optional[Union[ObjectRef, BaseEstimator, str]] = None,
                 custom_junior_config: Optional[dict] = None,
                 num_workers: Optional[int] = None,
                 subset: Optional[bool] = False,
                 env_kwargs: Optional[dict] = {},
                 rho_threshold: Optional[float] = 0.95):
        """
        The Senior requires multiple inputs for the initialization.

        After the init is complete, you can train the Senior or restore a checkpoint.
        Further, we enable a possible check of the underlying environment and model.

        Args:
            env_path: Path to the Grid2Op Environment. This has to be initialized within the methode.
            action_space_path: Either path to the actions or a list containing mutliple actions.
            model_path: The required model path loads the underling model for the senior and consist of
            one of either a Junior model or a Senior model.
            ckpt_save_path: Optional path, where the PPO should save its checkpoints. If not provided,
            scaler: If you want, you can pass a Scaler of Sklearn Model.
            Either a Sklearn Scaler or its ray ID, if the scaler is saved via ray.put().
            If scaler is provided, the environment will scale the observations based on scaler.transform()
            custom_junior_config: If the junior model is a model after the hyperparameter training, you
            need to pass the model configurations.
            ray will save them in the default directory ray_results.
            num_workers: You can configure the number of workers, based on your ray.init() configurations.
            If not specified, the PPO will used half of you CPU count.
            subset: Optional argument, whether the observations should be filtered when saved.
            The default version saves the observations according to obs.to_vect(), however if
            subset is set to True, then only the all observations regarding the lines, busses, generators and loads are
            selected.
            env_kwargs: Optional parameters for the Grid2Op environment that should be used when making the environment.
        """
        # ToDo optimize docstrings
        # Set default values:

        self.run_with_tf = run_with_tf
        self.ckpt_save_path = ckpt_save_path
        assert ray.is_initialized(), "Ray seems not to be initialized. Please use ray.init() prior to running" \
                                     "the Senior."

        if isinstance(scaler, (str,Path)):
            try:
                with open(scaler, "rb") as fp:  # Pickling
                    scaler = pickle.load(fp)
            except Exception as e:
                scaler = None
                logging.info(f"The scaler provided was either a path or a string. However, loading "
                             f"the scaler cause the following exception:{e}"
                             f"It will be set to None")

        if not rho_threshold:
            rho_threshold = 0.95

        self.env_config = {
            "action_space_path": action_space_path,
            "env_path": env_path,
            "action_threshold": rho_threshold,
            'subset': subset,
            'scaler': scaler,
            'topo': True,
            "env_kwargs": env_kwargs}

        self.model_config = None

        if self.run_with_tf:
            ModelCatalog.register_custom_model('Senior', Grid2OpCustomModelTF)
        else:
            ModelCatalog.register_custom_model('Senior', Grid2OpCustomModelTorch)

        if isinstance(custom_junior_config, (dict, str)):
            if isinstance(custom_junior_config, str):
                with open(custom_junior_config) as json_file:
                    custom_junior_config = json.load(json_file)

            self.model_config = {"model_path": model_path,
                                 "custom_config": custom_junior_config}
            self.__advanced_model = True
        else:

            self.model_config = {"model_path": model_path}
            self.__advanced_model = False

        # Testing of model and init the SeniorEnvRllib
        self.rllib_env: SeniorEnvRllib = None
        self.__test_env_and_model_config()

        # Now init PPO
        num_cpu = os.cpu_count()
        if not num_workers:
            num_workers = num_cpu // 2

        if self.run_with_tf:
            self.ppo_config = (
                PPOConfig().environment(env=SeniorEnvRllib, env_config=self.env_config)
                .rollouts(num_rollout_workers=num_workers)
                .framework("tf2")
                .training(model={"custom_model": "Senior",
                                 "custom_model_config": self.model_config})
                .evaluation(evaluation_num_workers=1))
        else:
            self.ppo_config = (
                PPOConfig().environment(env=SeniorEnvRllib, env_config=self.env_config)
                .rollouts(num_rollout_workers=num_workers)
                .framework("torch")
                .training(model={"custom_model": "Senior",
                                 "custom_model_config": self.model_config})
                .evaluation(evaluation_num_workers=1))


        self.ppo = self.ppo_config.build()

    def train(self, iterations: int = 1) -> dict:
        """ Train the Senior with the underlying PPO agent.

        Args:
            iterations: Number of Iterations for the PPO

        Returns: rllib output
        """
        out = None
        for i in range(iterations):
            out = self.ppo.train()

            # For every 5 steps, we save the current checkpoint:
            if i % 5 == 0:
                self.ppo.save(checkpoint_dir=self.ckpt_save_path)

        # Now save final checkpoint
        save_path = self.ppo.save(checkpoint_dir=self.ckpt_save_path)

        logging.info("An Algorithm checkpoint has been created inside directory: "
                     f"'{save_path}'.")

        return out

    def restore(self, path: Optional[Union[str, Path]]) -> None:
        """
        Restores the provided checkpoint. Alternatively you can also use the Algorithm.from_checkpoint()
        for this.
        Args:
            path: Path to checkpoint.

        Returns: None

        """
        self.ppo.restore(path)
        logging.info(f"Restored path: {path} ")


    def save_to_model(self, path: Optional[Union[str, Path]] = "."):
        """
        Saving the model for the final Agent. This model is saved as a TensorFlow model and
        can be loaded by the MyAgent method of the CurriculumAgent.

        Args:
            path: Path, where to save the model.

        Returns:

        """
        self.ppo.export_policy_model(path)
        logging.info(f"The MyAgent model is saved under {path}")

    def get_my_agent(self, path: Optional[Union[str, Path]] = ".") -> MyAgent:
        """
        Saves the Senior model and returns the final MyAgent model.

        Returns: MyAgent model with the respective action sets of the Senior

        """
        # First Save the PPO:
        self.save_to_model(path)

        # Load the my_agent:
        if self.run_with_tf:
            agent = MyAgent(
                action_space=self.rllib_env.single_env.action_space,
                model_path=path,
                action_space_path=self.env_config["action_space_path"],
                scaler=self.env_config["scaler"],
                best_action_threshold=self.env_config["action_threshold"],
                topo=self.env_config["topo"],
                subset=self.env_config["subset"],
                run_with_tf=True
            )
        else:
            agent = MyAgent(
                action_space=self.rllib_env.single_env.action_space,
                model_path=path,
                action_space_path=self.env_config["action_space_path"],
                scaler=self.env_config["scaler"],
                best_action_threshold=self.env_config["action_threshold"],
                topo=self.env_config["topo"],
                subset=self.env_config["subset"],
                run_with_tf=False
            )

        return agent

    def __test_env_and_model_config(self) -> None:
        """ This method tests, whether the inputs of the senior are sufficient enough and
        if everything works. This also means running the rllib method check_env()

        Returns: Nothing. The method should complete or else you have a problem.
        """
        # Create the senior_env_rllib:
        logging.info("Init of SeniorEnvRllib and testing one simple execution")
        self.rllib_env = SeniorEnvRllib(self.env_config)

        assert isinstance(self.rllib_env, SeniorEnvRllib), "The initialization of the SeniorEnvRllib failed!"
        # Run Environment:
        obs, _ = self.rllib_env.reset()
        term = False
        trunc = False
        while term is False and trunc is False:
            act = random.randrange(self.rllib_env.action_space.n)
            _, _, term, trunc, _ = self.rllib_env.step(act)
        obs, _ = self.rllib_env.reset()

        logging.info("The SeniorEnvRllib environment seems to run. Next, we check ray")
        check_env(self.rllib_env)
        logging.info("The SeniorEnvRllib check completed. ")

        # Now the model
        logging.info("Analyzing the Model configuration.")

        if self.run_with_tf:
            # First TF Model
            logging.info("First loading the TensorFlow model")
            model = tf.keras.models.load_model(self.model_config["model_path"])
            model.compile()
            logging.info("TF Model Import works")
        else:
            # First Torch Model
            logging.info("First loading the PyTorch model")
            model = Grid2OpCustomModelTorch(obs_space=self.rllib_env.observation_space,
                                           action_space=self.rllib_env.action_space,
                                           num_outputs=self.rllib_env.action_space.n,
                                           model_config={},
                                           model_path=self.model_config["model_path"],
                                           custom_config=self.model_config.get("custom_config", None),
                                           name="Junior")
            model._params_copy(self.model_config["model_path"])
            logging.info("PyTorch Import works")

        # Now the RLlib Model:
        if self.run_with_tf:
            if self.__advanced_model:
                model = Grid2OpCustomModelTF(obs_space=self.rllib_env.observation_space,
                                           action_space=self.rllib_env.action_space,
                                           num_outputs=self.rllib_env.action_space.n,
                                           model_config={},
                                           model_path=self.model_config["model_path"],
                                           custom_config=self.model_config["custom_config"],
                                           name="Junior")
            else:
                model = Grid2OpCustomModelTF(obs_space=self.rllib_env.observation_space,
                                           action_space=self.rllib_env.action_space,
                                           num_outputs=self.rllib_env.action_space.n,
                                           model_config={},
                                           model_path=self.model_config["model_path"],
                                           name="Junior")
        else:
            if self.__advanced_model:
                model = Grid2OpCustomModelTorch(obs_space=self.rllib_env.observation_space,
                                           action_space=self.rllib_env.action_space,
                                           num_outputs=self.rllib_env.action_space.n,
                                           model_config={},
                                           model_path=self.model_config["model_path"],
                                           custom_config=self.model_config["custom_config"],
                                           name="Junior")
            else:
                model = Grid2OpCustomModelTorch(obs_space=self.rllib_env.observation_space,
                                           action_space=self.rllib_env.action_space,
                                           num_outputs=self.rllib_env.action_space.n,
                                           model_config={},
                                           model_path=self.model_config["model_path"],
                                           name="Junior")

        logging.info("Init of model worked.")
        # Now Testing
        if self.run_with_tf:
            obs = {"obs": obs.reshape(1, -1)}
        else:
            obs = {"obs": torch.from_numpy(obs.reshape(1, -1))}
        assert model.forward(input_dict=obs, state=1, seq_lens=None), "Error, the model was not able to pass values!"
        logging.info("Model seems to be working")
        logging.info("All testing completed. ")



"""This file constist of two custom models, which transfer the junior weights into the Rllib experiment.

"""
import pickle
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Tuple, List

import numpy as np
import pathlib
import tensorflow as tf
import torch
import torch.nn as nn
from gymnasium.spaces import Discrete, Box
from ray.rllib.models.tf.tf_modelv2 import TFModelV2
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from torch.nn import Parameter
import lightning as L



# Import torch modules:

# from ray.rllib.utils.framework import try_import_torch

# torch, nn = try_import_torch()


class Grid2OpCustomModelTF(TFModelV2):
    """
    A custom TensorFlow model for policy gradient algorithms in the Grid2Op environment.

    This class defines a custom model architecture, with the ability to load pretrained models such as the Junior or
    Senior models, and is used within the RLlib framework for training in the Grid2Op environment.
    """

    def __init__(
            self,
            obs_space: Box,
            action_space: Discrete,
            num_outputs: int,
            model_config: dict,
            name: str,
            **customized_model_kwargs
    ):
        """
        Constructor for the custom TensorFlow model.

        This method initializes the model using the provided observation space, action space, and model configuration.
        It supports loading pretrained Junior or Senior models and allows customization of layers, activation functions,
        and initializers.

        Args:
            obs_space (Box): The observation space provided by RLlib.
            action_space (Discrete): The action space provided by RLlib.
            num_outputs (int): Number of output nodes, corresponding to the size of the action space.
            model_config (dict): Model configuration dictionary from RLlib.
            name (str): Name of the model.
            customized_model_kwargs (dict): Custom configuration for the model, including layer sizes, activation
                functions, and initializers.

        Returns:
            None.
        """
        super(Grid2OpCustomModelTF, self).__init__(obs_space, action_space, num_outputs, model_config, name)
        if "custom_config" in customized_model_kwargs.keys():
            cconfig = customized_model_kwargs["custom_config"]
        else:
            cconfig = {}
        layer_size, initializer, activation = self._extract_config(cconfig)

        # Now Init model:
        self.inputs = tf.keras.layers.Input(shape=obs_space.shape, name="observations")

        layer1 = tf.keras.layers.Dense(
            layer_size[0], name="layer_1", activation=activation,
            kernel_initializer=initializer[0])(
            self.inputs
        )
        layer2 = tf.keras.layers.Dense(layer_size[1], name="layer_2",
                                       activation=activation,
                                       kernel_initializer=initializer[1])(
            layer1
        )
        layer3 = tf.keras.layers.Dense(
            layer_size[2], name="layer_3",
            activation=activation,
            kernel_initializer=initializer[2]
        )(layer2)
        layer4 = tf.keras.layers.Dense(
            layer_size[3], name="layer_4",
            activation=activation,
            kernel_initializer=initializer[3]
        )(layer3)
        act_layer = tf.keras.layers.Dense(
            num_outputs, name="action_out",
            activation=None,
            kernel_initializer=initializer[4]
        )(layer4)
        val_hidden_layer = tf.keras.layers.Dense(
            action_space.n, name="layer_val_hidden",
            activation=activation,
            kernel_initializer=initializer[5]
        )(layer4)
        val_layer = tf.keras.layers.Dense(1, name="value_out",
                                          activation=None,
                                          kernel_initializer=initializer[6])(
            val_hidden_layer
        )

        self.base_model = tf.keras.Model(self.inputs, [act_layer, val_layer])

        path_to_junior = customized_model_kwargs["model_path"]

        self._params_copy(path=path_to_junior)

    def _extract_config(self, config) -> Tuple[List, tf.keras.initializers.Initializer, tf.keras.layers.Layer]:
        """
        Extracts model configuration parameters, including layer sizes, initializers, and activation functions.

        Args:
            config (dict): Dictionary containing model configuration parameters.

        Returns:
            Tuple[List, tf.keras.initializers.Initializer, tf.keras.layers.Layer]: A tuple containing:
                - A list of layer sizes.
                - A list of initializers for each layer.
                - The activation function to be used in the model.
        """

        layer_size = [1000, 1000, 1000, 1000]
        initializer = [tf.keras.initializers.Orthogonal(seed=np.random.choice(100000)) for _ in range(7)]
        activation = tf.keras.layers.ReLU()

        if not any([k in config.keys() for k in ["activation", "initializer", "layer1", "layer2", "layer3", "layer4"]]):
            warnings.warn("The custom dictionary did not have the correct keys. Using default model.")
            return layer_size, initializer, activation

        # Activation options:
        activation_option = {"leaky_relu": tf.keras.layers.LeakyReLU(),
                             "relu": tf.keras.layers.ReLU(),
                             "elu": tf.keras.layers.ELU(),
                             "tanh": tf.keras.activations.tanh}

        # Initializer Options with multiple seeds:
        initializer_option = {"O": tf.keras.initializers.Orthogonal(seed=np.random.choice(100000)),
                              "RN": tf.keras.initializers.RandomNormal(seed=np.random.choice(100000)),
                              "RU": tf.keras.initializers.RandomUniform(seed=np.random.choice(100000)),
                              "Z": tf.keras.initializers.Zeros()}

        activation = activation_option[config.get("activation", "relu")]

        initializer = [initializer_option[config.get("initializer", "O")] for _ in range(7)]

        layer_size = [int(np.round(config.get(l, 1000))) for l in ["layer1", "layer2", "layer3", "layer4"]]

        return layer_size, initializer, activation

    def _params_copy(self, path: Path):
        """Private Method from PPO code. Overwriting the weights of the rllib model by the Junior model.

        This method does one of two things:
        1. Used to copy the weights of the Junior model onto the model of the Rllib custom model.
        2. If the model is already a PPO model, it just copies it.

        Note that for the second option, the custom config does not have an effect.

        Args:
             path (Path): The file path to the saved model checkpoint.
        Returns:
            None.

        """
        model = tf.keras.models.load_model(path)

        if isinstance(model, tf.keras.models.Sequential):
            self.base_model.layers[1].set_weights(model.layers[0].get_weights())
            self.base_model.layers[2].set_weights(model.layers[1].get_weights())
            self.base_model.layers[3].set_weights(model.layers[2].get_weights())
            self.base_model.layers[4].set_weights(model.layers[4].get_weights())
            self.base_model.layers[5].set_weights((*map(lambda x: x / 5, model.layers[6].get_weights()),))
        else:
            model.compile()
            self.base_model = model

    def forward(self, input_dict, state, seq_lens):
        """
        Performs a forward pass through the model.

        Args:
            input_dict (dict): A dictionary containing the observation data from RLlib.
            state: The state input (not used in this model).
            seq_lens: Sequence lengths (not used in this model).

        Returns:
             The action logits and the state.
        """
        input_dict["obs"]
        model_out, self._value_out = self.base_model(input_dict["obs"])
        return model_out, state

    def value_function(self):
        """
        Returns the value function output of the model.

        This is used by RLlib to compute the value function during training.

        Returns:
            tf.Tensor: The value function output.
        """
        return tf.reshape(self._value_out, [-1])

    def metrics(self):
        """
        Returns custom metrics for the model.

        This method provides a placeholder metric that can be extended or customized depending on the model's
        requirements. In this case, it returns a constant value as a demonstration.

        Returns:
            dict: A dictionary containing the custom metric.
        """
        return {"foo": tf.constant(42.0)}


class Grid2OpCustomModelTorch(TorchModelV2, nn.Module):
    """
    A custom PyTorch model for policy gradient algorithms in the Grid2Op environment.

    This class defines a custom neural network architecture, and supports loading pretrained models, including
    the Junior and Senior models, for use within the RLlib framework in the Grid2Op environment.
    """
    def __init__(self, obs_space, action_space, num_outputs, model_config, name, **customized_model_kwargs):
        TorchModelV2.__init__(
            self, obs_space, action_space, num_outputs, model_config, name
        )
        """
        Constructor for the custom PyTorch model.

        This method initializes the custom model using the observation space, action space, and model configuration 
        from RLlib. It supports layer customization and loading pretrained models, such as Junior models.

        Args:
            obs_space (Box): The observation space provided by RLlib.
            action_space (Discrete): The action space provided by RLlib.
            num_outputs (int): Number of output nodes, corresponding to the action space.
            model_config (dict): Configuration dictionary for the model.
            name (str): Name of the model.
            customized_model_kwargs (dict): Custom model parameters including layer sizes, activation functions, 
                and initializers.

        Returns:
            None.
        """
        nn.Module.__init__(self)
        self.obs_space = obs_space
        self.action_space = action_space
        self.num_outputs = num_outputs
        self.model_config = model_config
        self.name = name

        if "custom_config" in customized_model_kwargs.keys() and customized_model_kwargs["custom_config"] is not None:
            cconfig = customized_model_kwargs["custom_config"]
        else:
            cconfig = {}
        layer_size, initializer, activation = self._extract_config(cconfig)

        # Define layers
        self.layer1 = nn.Linear(obs_space.shape[0], layer_size[0])
        self.layer2 = nn.Linear(layer_size[0], layer_size[1])
        self.layer3 = nn.Linear(layer_size[1], layer_size[2])
        self.layer4 = nn.Linear(layer_size[2], layer_size[3])
        self.act_layer = nn.Linear(layer_size[3], num_outputs)
        self.val_hidden_layer = nn.Linear(layer_size[3], action_space.n)
        self.val_layer = nn.Linear(action_space.n, 1)

        # Set activation and initializer
        self.activation1 = deepcopy(activation)
        self.activation2 = deepcopy(activation)
        self.activation3 = deepcopy(activation)
        self.activation4 = deepcopy(activation)
        self.activation5 = activation
        self._apply_initializer(initializer)
        self.softmax = nn.Softmax(dim=1)

        # Load model if path provided
        path_to_junior = customized_model_kwargs.get("model_path", None)
        if path_to_junior:
            self._params_copy(path=path_to_junior)

    def forward(self, input_dict, state=None, seq_lens=None):
        """
        Performs a forward pass through the model.

        This method processes the input observation through the layers and computes the action logits and value output.

        Args:
            input_dict (dict): A dictionary containing the observation data from RLlib.
            state: The state input (not used in this model).
            seq_lens: Sequence lengths (not used in this model).

        Returns:
            Tuple[torch.Tensor, None]: The action logits and the state (None in this case).
        """
        # Extract the observations from the input dictionary
        if isinstance(input_dict["obs"], np.ndarray):
            input = torch.from_numpy(input_dict["obs"])
        else:
            input = input_dict["obs"]

        x = input
        # Apply the layers and activations
        x = self.activation1(self.layer1(x))
        x = self.activation2(self.layer2(x))
        x = self.activation3(self.layer3(x))
        x = self.activation4(self.layer4(x))

        # Action and value outputs
        action = self.act_layer(x)
        val_hidden = self.activation5(self.val_hidden_layer(x))
        value = self.val_layer(val_hidden)

        # Store value for value function calls later
        self._value_out = value.squeeze(-1)

        # Apply softmax to action
        action = self.softmax(action)

        # Return action logits and state
        return action, state

    def _extract_config(self, config) -> Tuple[List, any, any]:
        """
        Extracts model configuration parameters, including layer sizes, activation functions, and initializers.

        Args:
            config (dict): Configuration dictionary containing model parameters.

        Returns:
            Tuple[List, any, any]: A tuple containing:
                - A list of layer sizes.
                - The initializer for the layers.
                - The activation function to be used in the model.
        """
        layer_size = [1000, 1000, 1000, 1000]
        initializer = nn.init.orthogonal_
        activation = nn.ReLU()

        if not any([k in config.keys() for k in ["activation", "initializer", "layer1", "layer2", "layer3", "layer4"]]):
            warnings.warn("The custom dictionary did not have the correct keys. Using default model.")
            return layer_size, initializer, activation

        activation_option = {"leaky_relu": nn.LeakyReLU(inplace=True),
                             "relu": nn.ReLU(inplace=True),
                             "elu": nn.ELU(inplace=True)}
        initializer_option = {"O": nn.init.orthogonal_,
                              "RN": nn.init.normal_,
                              "RU": nn.init.uniform_,
                              "Z": nn.init.zeros_}

        if "activation" in config:
            activation = activation_option[config["activation"]]

        if "initializer" in config:
            initializer = initializer_option[config["initializer"]]

        if all([l in config for l in ["layer1", "layer2", "layer3", "layer4"]]):
            layer_size = [int(round(config[l])) for l in ["layer1", "layer2", "layer3", "layer4"]]

        return layer_size, initializer, activation

    def _apply_initializer(self, initializer):
        """
        Applies the specified initializer to the model layers.

        Args:
            initializer (Callable): The initializer function for the model weights.

        Returns:
            None.
        """
        for m in self.modules():
            if isinstance(m, nn.Linear):
                initializer(m.weight)

    def load_my_state_dict(self, state_dict):
        """
        Loads the model's state dictionary, mapping the given state dictionary to the current model's parameters.

        This method adjusts the keys of the provided `state_dict` to match the current model's state dictionary
        and copies the weights from the provided dictionary into the corresponding parameters of the model.

        Args:
            state_dict (dict): The state dictionary containing the model weights. This can be a saved checkpoint
                from a pretrained model.

        Returns:
            None.
        """
        own_state = self.state_dict()

        for name, param in state_dict.items():
            name_corr = name
            if name.startswith("model."):
                name_corr = name[len("model."):]
            if name_corr not in own_state:
                continue
            if isinstance(param, Parameter):
                param = param.data
            own_state[name_corr].copy_(param)

    def _params_copy(self, path: Path):
        """
         Copies the parameters from a pretrained model to the custom PyTorch model.

         This method loads the weights of a pretrained Junior or Senior model from the specified path
         and applies them to the custom PyTorch model.

         Args:
             path (Path): The file path to the saved model checkpoint.

         Returns:
             None.
         """
        checkpoint = torch.load(path)
        try:
            state_dict = checkpoint.get('state_dict', checkpoint)
        except AttributeError:
            print("Loading the Senior model")
            state_dict = checkpoint.state_dict()

        self.load_my_state_dict(state_dict)

    def value_function(self):
        """
        Returns the value function output of the model.

        This is used by RLlib to compute the value function during training.

        Returns:
            The value function output.
        """
        return self._value_out



import logging
from pathlib import Path
from typing import Union, Optional, Tuple, List

import nni
import numpy as np

from collections import ChainMap

import tensorflow as tf
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint, Callback, EarlyStopping
from tensorflow.keras.optimizers.schedules import LearningRateSchedule
from tensorflow.keras.initializers import Initializer

from typing import Union, Optional, TypedDict, Tuple, List

from curriculumagent.common.utilities import map_actions


class JuniorParamTF(TypedDict):
    """TypeDict Class with following Attributes:
    action_space_path: Either path to the actions or a list containing mutliple actions.
    data_path: Path of the Grid2Op environment
    action_threshold: Between 0 and 1
    subset: Should the obs.to_vect be filtered similar to the original Agent. If True,
            the obs.to_vect is filtered based on predefined values. If False, all values are considered.
            Alternatively, one can submit the once own values.
            testing: Indicator, whether the underlying Grid2Op Environment should be started in testing mode or not
    """

    activation: str
    learning_rate: Optional[Union[float, LearningRateSchedule]]
    layer1: int
    layer2: int
    layer3: int
    layer4: int
    batchsize: int
    dropout1: float
    dropout2: float
    epochs: int
    initializer: Initializer


class JuniorTF:
    def __init__(
            self,
            action_space_file: Union[Path, List[Path]],
            config: JuniorParamTF =None,
            seed: Optional[int] = None,
            run_nni: bool = False
    ):
        """Constructor of the Junior simple model or model with more flexible parameters. This Junior can also be used
        for the hyperparameter search with tune ore with NNI.

        Except setting all variables, the init additionally requires the size of the train set and optionally
        the number of epochs.

        Note:
            Pass epochs, learning_rate, batchsize, layer_size in the config.

        Args:
            action_space_file: Action Space file that was used for the Tutor training. This is needed to extract the
            correct shape of the Junior model.
            config: Dictionary containing the correct input for the hyperparameters.
            seed: Optional Seed to reproduce results.
            run_nni: Whether NNI is used. If True, then a specific callback is added.

        Returns:
            None.
        """

        # Set self actions to a list to iterate for later
        if config is None:
            config = {}

        if isinstance(action_space_file, Path):
            assert action_space_file.is_file()
            self.actions = np.load(str(Path(action_space_file)))

        elif isinstance(action_space_file, list):
            for act_path in action_space_file:
                assert act_path.is_file()
            self.actions = np.concatenate([np.load(str(act_path)) for act_path in action_space_file], axis=0)

        else:
            self.actions = np.load(action_space_file)

        self.num_actions = len(self.actions)

        self.config = config

        layer_size, initializer, activation = self._extract_config(config)

        self.lr = config.get("learning_rate", 5e-4)
        self.batch_size = config.get("batchsize", 256)
        self.epochs = config.get("epochs", 1000)

        self.activation = activation
        self.layer_size = layer_size
        self.lf = tf.keras.losses.SparseCategoricalCrossentropy()
        self.initializer = tf.keras.initializers.Orthogonal(seed=seed)

        # Seed:
        self.seed = seed
        if self.seed:
            np.random.seed(self.seed)
            tf.random.set_seed(self.seed)



        # Init Model: (either simple or advanced)
        self.model = self._build_model()
        if run_nni:
            self.callback = [SendMetrics()]
        else:
            self.callback = []

    def _build_model(self) -> tf.keras.models.Sequential:
        """Build and return the junior network as a keras model.

        Args:
            None.

        Returns:
            Compiled Keras model.

        """
        if not self.config:
            # build standart juniour model

            model = tf.keras.models.Sequential(
                [
                    tf.keras.layers.Dense(units=1000, activation=self.activation, kernel_initializer=self.initializer),
                    tf.keras.layers.Dense(units=1000, activation=self.activation, kernel_initializer=self.initializer),
                    tf.keras.layers.Dense(units=1000, activation=self.activation, kernel_initializer=self.initializer),
                    tf.keras.layers.Dropout(0.25),
                    tf.keras.layers.Dense(units=1000, activation=self.activation, kernel_initializer=self.initializer),
                    tf.keras.layers.Dropout(0.25),
                    tf.keras.layers.Dense(self.num_actions, activation="softmax"),
                ]
            )

        else:
            # Build model based on hyperparameter:

            # Layer1:
            model_structure = [
                tf.keras.layers.Dense(self.layer_size[0], activation=self.activation, kernel_initializer=self.initializer)
            ]
            # Layer2:
            if self.config["layer2"] != 0:
                model_structure += [
                    tf.keras.layers.Dense(self.layer_size[1], activation=self.activation,
                                          kernel_initializer=self.initializer)
                ]
            # Layer3:
            if self.config["layer3"] != 0:
                model_structure += [
                    tf.keras.layers.Dense(self.layer_size[2], activation=self.activation,
                                          kernel_initializer=self.initializer)
                ]
            # Dropout 1:
            if self.config["dropout1"] != 0.0:
                model_structure += [tf.keras.layers.Dropout(self.config["dropout1"])]

            # Layer 4:
            if self.config["layer4"] != 0:
                model_structure += [
                    tf.keras.layers.Dense(self.layer_size[3], activation=self.activation,
                                          kernel_initializer=self.initializer)
                ]

            # Dropout 2:
            if self.config["dropout2"] != 0.0:
                model_structure += [tf.keras.layers.Dropout(self.config["dropout2"])]

            model_structure += [tf.keras.layers.Dense(self.num_actions, activation="softmax")]
            model = tf.keras.models.Sequential(model_structure)

        model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=self.lr),
                      loss=self.lf,
                      metrics=["accuracy"])
        logging.info(model.summary)
        return model

    def _extract_config(self, config) -> Tuple[List, tf.keras.initializers.Initializer, tf.keras.layers.Layer]:
        """
        Method to extract the hyperparameters of the config

        Args:
            config:

        Returns: layers, initializer and activation methode.

        """
        # Default values:
        layer_size = [1000, 1000, 1000, 1000]
        initializer = tf.keras.initializers.Orthogonal()
        activation = tf.keras.layers.ReLU()

        if not any([k in config.keys() for k in ["activation", "initializer", "layer1",
                                                 "layer2", "layer3", "layer4"]]):
            import warnings
            warnings.warn("The custom dictionary had not the correct keys. Revert to default model.")
            return layer_size, initializer, activation

        # Activation options:
        activation_option = {"leaky_relu": tf.keras.layers.LeakyReLU(),
                             "relu": tf.keras.layers.ReLU(),
                             "elu": tf.keras.layers.ELU(),
                             }
        # Initializer Options
        initializer_option = {"O": tf.keras.initializers.Orthogonal(),
                              "RN": tf.keras.initializers.RandomNormal(),
                              "RU": tf.keras.initializers.RandomUniform(),
                              "Z": tf.keras.initializers.Zeros()}

        activation = activation_option[config.get("activation", "relu")]
        initializer = initializer_option[config.get("initializer", "O")]

        if all([l in config.keys() for l in ["layer1", "layer2", "layer3", "layer4"]]):
            layer_size = [int(np.round(config[l])) for l in ["layer1", "layer2", "layer3", "layer4"]]

        return layer_size, initializer, activation

    def train(
            self,
            x_train: np.ndarray,
            y_train: np.ndarray,
            x_validate: np.ndarray,
            y_validate: np.ndarray,
            log_dir: Optional[Union[str, Path]] = None,
            ckpt_dir: Optional[Union[str, Path]] = None,
            patience: Optional[int] = None,
            epochs: Optional[int] = None
    ) -> tf.keras.callbacks.History:
        """Train the junior model for given number of epochs.

        This method builds callbacks for the training and then executes the keras .fit() method
        to train the Junior model on the x_train and y_train data. Validation is recorded as well.

        Args:
            log_dir: Directory for tensorboard callback.
            ckpt_dir: Directory for checkpoint callback.
            x_train: Training data containing the grid observations.
            y_train: Training actions of the tutor.
            x_validate: Validation data containing the grid observations.
            y_validate: Validation actions of the tutor.
            epochs: Number of epochs for the training.
            patience: Optional early stopping criterion.

        Returns:
            Returns training history.

        """
        callbacks = self.callback

        if log_dir is not None:
            tensorboard_callback = TensorBoard(log_dir=log_dir, write_graph=False)
            callbacks += [tensorboard_callback]
        if isinstance(ckpt_dir, (Path, str)):
            if isinstance(ckpt_dir, str):
                ckpt_path = ckpt_dir + "/" + "ckpt_{epoch}"
            else:
                ckpt_path = ckpt_dir / "ckpt_{epoch}"

            cp_callback = ModelCheckpoint(filepath=str(ckpt_path),
                                          save_weights_only=False,
                                          save_freq=10, verbose=0)
            callbacks += [cp_callback]



        if patience is not None:
            early_stopping = EarlyStopping(
                monitor="val_loss",
                patience=patience,
                verbose=1,
                mode="auto",
                restore_best_weights=True,
            )
            callbacks += [early_stopping]

        history = self.model.fit(
            x=x_train,
            y=y_train,
            epochs=epochs or self.epochs,
            validation_data=(x_validate, y_validate),
            batch_size=self.batch_size,
            callbacks=callbacks,
            verbose=1
        )
        return history

    def test(self, x: np.ndarray, y: np.ndarray, save_path: Optional[Path] = None) -> dict:
        """Test the Junior model with input dataset x and targets/actions y.

        The method predicts based on the input x and then computes a ranking, regarding the
        accuracy on the actions.

        The ranking collects, if the action of the tutor was within the 1-20 actions.

        Args:
            x: Input with Tutor observation for the prediction.
            y: Action of the tutor to validate with the prediction.
            save_path: Optional path where the weights of the model are saved.
                If needed, the weights are loaded by model.load_weights(save_path).

        Returns:
            The dictionary that contains the top values.

        """
        if isinstance(save_path, Path):
            self.model = tf.keras.models.load_model(save_path)
            logging.info(f"Imported model from {save_path}")

        a_pred = self.model.predict(x, verbose=1)
        top_n = []
        for i in range(a_pred.shape[0]):
            top_n.append(a_pred[i, :].argsort()[-20:])

        # Added accuracy to record the prediction performance
        accuracy = {}

        for n in range(1, 21):
            correct = 0
            for i in range(a_pred.shape[0]):
                if y[i, 0] in top_n[i][-n:]:
                    correct += 1

            acc = correct / a_pred.shape[0] * 100
            logging.info(f"accuracy of top-{n} is {acc}")

            accuracy["accuracy of top-%d" % n] = correct / a_pred.shape[0] * 100
        return accuracy





class SendMetrics(tf.keras.callbacks.Callback):
    """ Keras callback to send metrics to NNI framework
    """

    def on_epoch_end(self, epoch, logs=None):
        """Keras callback to send the intermediate result to NNI

        Args:
            epoch: Epoch of training
            logs: Log input

        Returns: None, reports the intermediate result to NNI

        """
        # TensorFlow 2.0 API reference claims the key is `val_acc`, but in fact it's `val_accuracy`
        if logs is None:
            logs = {}
        if "val_acc" in logs:
            nni.report_intermediate_result(logs["val_acc"])
        else:
            nni.report_intermediate_result(logs["val_accuracy"])


"""This file runs the tuple and triple search defined in :mod:`topology_action_search` and saves the collected
teacher_experience in a csv file for further processing using :mod:`generate_action_space`.

"""
import logging
import os
import time
from multiprocessing import Pool
from pathlib import Path
from typing import Optional, List, Tuple, Union

import defopt
import grid2op
import numpy as np
from grid2op.Action import BaseAction
from grid2op.Environment import BaseEnv
from grid2op.Parameters import Parameters

from curriculumagent.teacher.submodule.common import save_sample_new
from curriculumagent.common.utilities import find_best_line_to_reconnect
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction
from curriculumagent.teacher.submodule.topology_action_search import topology_search_tuples, topology_search_triples


def collect_multi_experience(
        save_path: Path,
        env_name_path: Union[str, Path],
        chronics_path: Optional[str],
        use_triple_search: bool,
        seed: int,
        rho_acting_threshold: float = 0.925,
        top_k: int = 125,
        chronic_limit: int = None,
):
    """Collect general teacher_experience by running a greedy search whenever the agent encounters a high power load.
    It also tries to combine actions to tuple actions.

    Args:
        save_path: The path to save the teacher_experience to.
        env_name_path: The name or path of the environment used.
        chronics_path: The path to the chronics of the environment.
        use_triple_search: Whether to use triple search instead of tuple search.
        seed: The seed of the environment.
        rho_acting_threshold: When the environments max_rho gues above this threshold the agent begins to act.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        chronic_limit: How many chronics to run. Runs forever if set to None.

    Returns:
        None.
    """
    if isinstance(env_name_path, Path):
        env_name_path = str(env_name_path)

    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
        if chronics_path:
            # Use env_name_path as path together with chronics_path
            env: BaseEnv = grid2op.make(dataset=env_name_path, chronics_path=chronics_path, backend=backend)
        else:
            # Use env_name_path as name
            env: BaseEnv = grid2op.make(dataset=env_name_path, backend=backend)
    except Exception as e:  # noqa
        logging.warning("Not using lightsim2grid! Operation will be slow!")
        raise e

    action_space = env.action_space

    params: Parameters = env.parameters
    params.MAX_SUB_CHANGED = 99
    params.MAX_LINE_STATUS_CHANGED = 99
    env.change_parameters(params)

    # Load action sets from files used by search algorithm
    def load_compressed_actionspace(path: Path) -> List[BaseAction]:
        return [action_space.from_vect(row) for row in np.load(str(path))["actions"]]

    data_path: Path = Path(__file__).parent.parent.parent / "data"

    all_unitary_actions = load_compressed_actionspace(data_path / "all_unitary_actions.npz")
    if use_triple_search:
        best_tuple_actions = load_compressed_actionspace(data_path / "best_tuple_actions.npz")

        def topo_search(enviro: BaseEnv) -> List[Tuple[float, EncodedTopologyAction]]:
            return topology_search_triples(enviro, best_tuple_actions, all_unitary_actions, show_progress=True)

    else:
        all_unitary_actions_not16 = load_compressed_actionspace(data_path / "all_unitary_actions_not16.npz")
        best_unitary_actions = load_compressed_actionspace(data_path / "best_unitary_actions300.npz")

        def topo_search(enviro: BaseEnv) -> List[Tuple[float, EncodedTopologyAction]]:
            return topology_search_tuples(
                enviro, best_unitary_actions, all_unitary_actions, all_unitary_actions_not16, show_progress=True
            )

    # Setup chronics with deterministic shuffling
    np.random.seed(seed)
    env.chronics_handler.shuffle(shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
    num_chronics = len(os.listdir(chronics_path))
    if chronic_limit:
        num_chronics = min(num_chronics, chronic_limit)
    for _ in range(num_chronics):
        env.reset()  # env.reset() loads the next chronic
        dst_step = 0
        logging.info(f"Scenario to test is [{env.chronics_handler.get_name()}], start from step-{dst_step:d}... ...")
        env.fast_forward_chronics(dst_step)
        obs, done = env.get_obs(), False
        while not done:
            if obs.rho.max() >= rho_acting_threshold:
                best_actions = topo_search(env)
                if len(best_actions) > 0:
                    action = best_actions[0][1].to_action(env)
                    obs_, _, done, _ = env.step(action)
                    save_sample_new(save_path, best_actions, obs, obs_, top_k=top_k)
                    obs = obs_
                else:
                    action = env.action_space({})
                    action = find_best_line_to_reconnect(obs, action)
                    obs, _, done, _ = env.step(action)
            else:
                action = env.action_space({})
                action = find_best_line_to_reconnect(obs, action)
                obs, _, done, _ = env.step(action)
            dst_step += 1


def run_multi_teacher(
        save_path: Path,
        use_triple_search: bool,
        env_name_path: str = "l2rpn_neurips_2020_track1_small",
        rho_acting_threshold: float = 0.925,
        *,
        n_episodes: int = 100,
        top_k: int = 125,
        limit_chronics: int = None,
        jobs: int = os.cpu_count(),
        seed: int = 42,
) -> Path:
    """This teacher is trying to collect useful actions by doing topology search while running an
    environment when an overflow occurs. It also tries to combine actions to form tuple or triple actions.

    Args:
        save_path: Path to save the teacher_experience file to.
        use_triple_search: Execute the triple teacher instead of the tuple teacher.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_acting_threshold: When the environments max_rho goes above this threshold the agent begins to act.
        n_episodes: Number of episodes (run through all scenarios) to run.
        limit_chronics: Limit the number of chronics to run through.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: The random seed to use.

    Returns:
        Saving path.
    """

    env_path = Path(env_name_path)
    chronics_path = None
    if env_path.is_dir():
        env_name_path = env_path
        chronics_path = str(env_name_path / "chronics")

    start = time.time()
    tasks = []
    for episode in range(n_episodes):
        tasks.append(
            (
                save_path,
                env_name_path,
                chronics_path,
                use_triple_search,
                seed + episode,
                rho_acting_threshold,
                top_k,
                limit_chronics,
            )
        )

    logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers: {tasks}")
    # Sequential implementation, useful for debugging more easily
    if jobs == 1:
        for task in tasks:
            collect_multi_experience(*task)
    else:
        # Parallel implementation
        with Pool(jobs, maxtasksperchild=1) as p:
            p.starmap(collect_multi_experience, tasks)
    end = time.time()
    elapsed = end - start
    logging.info(f"Needed time: {elapsed}s")

    return save_path






"""This file is build to construct the environment of the Senior Agent to run the PPO with ray and rllib.
"""
import logging
import os
import pickle
from pathlib import Path
from typing import Union, TypedDict, Tuple, List, Optional, Dict

import grid2op
import gymnasium as gym
import numpy as np
import ray
from grid2op.Exceptions import UnknownEnv
from gymnasium.spaces import Discrete, Box
from lightsim2grid import LightSimBackend
from ray._raylet import ObjectRef
from sklearn.base import BaseEstimator

from curriculumagent.common.obs_converter import obs_to_vect
from curriculumagent.common.utilities import find_best_line_to_reconnect, split_and_execute_action, revert_topo, \
    set_bus_from_topo_vect
from curriculumagent.senior.rllib_execution.alternative_rewards import PPO_Reward


class RLlibEnvConfig(TypedDict):
    """
    A TypedDict class representing the configuration parameters for an RLlib environment in the Grid2Op framework.

    This configuration is used to initialize and configure the environment, including options for action spaces,
    scaling, and rewards.

    Attributes:
        action_space_path (Union[Path, List[Path]]): Either a path to the action space file or a list containing
            multiple action space files.
        env_path (Union[str, Path]): Path to the Grid2Op environment.
        action_threshold (float): A threshold value between 0 and 1 that defines the cutoff for selecting actions.
        filtered_obs (Union[bool, List]): If True, the `obs.to_vect` is filtered based on predefined values.
            If False, all values are considered. Alternatively, a custom subset of values can be provided.
        scaler (Optional[Union[ObjectRef, BaseEstimator, str]]): An optional scaler for normalizing observations.
            This can be a Sklearn scaler, a ray object reference if saved using `ray.put()`, or a string ID.
        topo (Optional[bool]): Whether to revert the topology in the environment or not.
        alternative_rew (Optional[grid2op.Reward.BaseReward]): An optional alternative reward function for the environment.
        env_kwargs (Optional[dict]): Additional arguments to be passed to the Grid2Op environment.
        seed (Optional[int]): Seed value for reproducibility. If provided, ensures the same environment conditions
            are reproducible.
    """
    action_space_path: Union[Path, List[Path]]
    env_path: Union[str, Path]
    action_threshold: float
    filtered_obs: Union[bool, List]
    scaler: Optional[Union[ObjectRef, BaseEstimator, str]]
    topo: Optional[bool]
    alternative_rew: Optional[grid2op.Reward.BaseReward]
    env_kwargs: Optional[dict] = {}
    seed: Optional[int]


class SeniorEnvRllib(gym.Env):
    """
    A custom environment class that combines the Grid2Op backend with the actions of the tutor agent for use
    in the RLlib framework.

    This environment interacts with the Grid2Op environment, making it suitable for RLlib-based training.
    Unlike the `GymEnv` provided by `grid2op.gym_compat`, this environment maintains direct access to the Grid2Op
    environment, offering more flexibility in interaction.

    Note:
        With the update to Ray>=2.4, the environment now uses the `gymnasium` package instead of `gym`. This change
        requires small modifications to the environment's structure and behavior.

    """

    def __init__(self, config: RLlibEnvConfig,
                 testing: bool = False):
        """
        Initializes the Senior environment with the Grid2Op backend and RLlib-compatible settings.

        This method sets up the environment using a configuration dictionary and, if possible, uses the
        `lightsim2grid` backend. The environment is structured based on the Gym format.

        Args:
            config (RLlibEnvConfig): The configuration dictionary containing the following keys and values:
                action_space_path (Union[Path, List[Path]]): Path to the action space file or a list of action files.
                env_path (Union[str, Path]): Path to the Grid2Op environment.
                action_threshold (float): A threshold value between 0 and 1 for selecting actions.
                subset (Union[bool, List]): If True, filters the `obs.to_vect` based on predefined values. If False,
                    all values are considered. Alternatively, custom values can be provided.
                testing (bool): Indicates whether the Grid2Op environment should be started in testing mode.
                scaler (Optional[Union[ObjectRef, BaseEstimator, str]]): Optional scaler for normalizing observations.
                alternative_rew (Optional[grid2op.Reward.BaseReward]): An alternative reward function for the environment.
                env_kwargs (Optional[dict]): Additional keyword arguments for configuring the Grid2Op environment.
            testing (bool, optional): If True, initializes the environment in testing mode. Defaults to False.

        Returns:
            None.
        """
        seed = config.get("seed",np.random.choice(10000))
        logging.info(f"Setting seed {seed}")
        self.seed = seed

        np.random.seed(seed)


        # Check if subsample of Obs:
        action_space_path = config["action_space_path"]
        action_threshold = config.get("action_threshold",1.0)

        self.subset = config.get("subset",False)

        data_path = str(config["env_path"])
        env_kwargs = config.get("env_kwargs", {})

        # Init environments: If testing is true, we test the test flag
        backend = LightSimBackend()
        try:
            if testing:
                env = grid2op.make(dataset=data_path, backend=backend,
                                   reward_class=PPO_Reward, test=True, **env_kwargs)
                logging.info("Test flag was set to true. Init Env in testing mode.")
            else:
                reward = config.get("alternative_rew",PPO_Reward)
                env = grid2op.make(dataset=data_path, backend=backend,
                                       reward_class=reward, test=False, **env_kwargs)
                logging.info("Init of Grid2Op Env works.")
            self.single_env = env.copy()
        except UnknownEnv as test_env:
            logging.info(f"Error testing was raised {test_env}. Assume testing, thus set test flag.")
            self.single_env = grid2op.make(dataset=data_path, test=True,
                                           backend=LightSimBackend(), reward_class=PPO_Reward)

        self.single_env.seed(self.seed)
        self.single_env.chronics_handler.shuffle(
            shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
        # Create multiple environments:
        self.obs_grid2op = self.single_env.reset()

        # Initialize the actions:
        if isinstance(action_space_path, Path):
            if action_space_path.is_file():
                logging.info(f"Action_space_path {action_space_path} is a file and will be loaded.")
                self.actions = np.load(str(action_space_path))
            elif action_space_path.is_dir():
                logging.info(f"Action_space_path {action_space_path} is a path. All available action files "
                             f" will be loaded.")
                all_action_files = [act for act in os.listdir(action_space_path) if "actions" in act and ".npy" in act]

                if not all_action_files:
                    raise FileNotFoundError("No actions files were found!")

                loaded_files = [np.load(str(action_space_path / act)) for act in all_action_files]
                self.actions = np.concatenate(loaded_files, axis=0)

        elif isinstance(action_space_path, list):
            logging.info(f"Action_space_path {action_space_path} is a list containing multiple actions.")
            loaded_files = [np.load(str(act_path)) for act_path in action_space_path]
            self.actions = np.concatenate(loaded_files, axis=0)
        else:
            raise ValueError(f"The action_space_path variable {action_space_path} does neither consist of a single "
                             f"action nor of a path where actions can be found.")

        # Define Action and Observation Space in Gym Format:
        self.action_space = Discrete(len(self.actions))

        if isinstance(self.subset, list):
            vect_shape = len(self.subset)
        elif self.subset:
            vect_shape = obs_to_vect(self.single_env.get_obs(), False).shape[0]
        else:
            vect_shape = self.single_env.observation_space.size()
        self.observation_space = Box(shape=(vect_shape,), high=np.inf, low=-np.inf)

        logging.info(f"Initialize Environment with the following action space{self.action_space}  and the following "
                     f"observations space{self.observation_space}")

        # check scaler:
        self.scaler = config.get("scaler", None)
        if self.scaler is not None:
            if isinstance(self.scaler, (str, Path)):
                with open(self.scaler, "rb") as fp:
                    self.scaler = pickle.load(fp)

        # Should we revert the topology ?
        self.topo = config.get("topo", False)

        # Check whether the actions are grid2op.Action.Baseactions in vector format or whether they are
        # actually topologies:
        topology_actions = not len(self.single_env.action_space().to_vect()) == self.actions.shape[1]
        self.topology_actions = config.get("topology_actions", topology_actions)

        # Initialize further arguments:
        assert 0.0 < action_threshold <= 1.0
        self.action_threshold = action_threshold

        self.step_in_env = 0
        self.passiv_reward = 0
        self.obs_rl = self.run_next_action()
        self.reward = None

        self.info: Dict = {}
        self.done = self.single_env.done
        # Added for gymnasium
        self.truncated, self.terminated = self.check_terminated_truncated()

    def run_next_action(self) -> np.ndarray:
        """
        Runs the environment until an action by the agent is required.

        In this function, the Grid2Op environment runs without agent intervention until the environment requires
        an action. The cumulative reward is calculated during this process, and the observation is returned when
        the agent is needed to take action.

        Returns:
            np.ndarray: The observation vector after the environment requires an action from the agent.
        """
        # init values
        self.passiv_reward = 0
        continue_running = True
        cur_obs = self.single_env.get_obs()
        cumulated_rew = 0
        # Run through the env and do nothing.
        # The reward is collected until the agent is done.
        # If done, there is an extra reward.
        while continue_running:
            if cur_obs.rho.max() < self.action_threshold:
                if self.topo:
                    action_array = revert_topo(self.single_env.action_space, cur_obs)
                    default_action = self.single_env.action_space.from_vect(action_array)
                else:
                    default_action = self.single_env.action_space({})

                action = find_best_line_to_reconnect(obs=cur_obs,
                                                     original_action=default_action)
                # Execute in env:
                cur_obs, reward, _, self.info = self.single_env.step(action)
                cumulated_rew += reward
                self.step_in_env += 1

                if self.single_env.done:
                    continue_running = False

            else:
                continue_running = False
        self.obs_grid2op = cur_obs
        self.passiv_reward = cumulated_rew

        # Select subset if wanted
        if isinstance(self.subset, list):
            obs_rl = self.obs_grid2op.to_vect()[self.subset]
        elif self.subset:
            obs_rl = obs_to_vect(self.obs_grid2op, False)
        else:
            obs_rl = self.obs_grid2op.to_vect()

        # Select scaler if wanted
        if self.scaler:
            if isinstance(self.scaler, BaseEstimator):
                obs_rl = self.scaler.transform(obs_rl.reshape(1, -1)).reshape(-1, )
            elif isinstance(self.scaler, ObjectRef):
                obs_rl = ray.get(self.scaler).transform(obs_rl.reshape(1, -1)).reshape(-1, )

        return obs_rl

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, Union[float, int], bool, bool, dict]:
        """
        Executes the agent's action in the environment and computes the reward.

        This method performs the action specified by the agent and accumulates rewards for passive actions taken
        by the environment. It checks if multiple steps are required and handles the transition based on the
        Grid2Op environment's conditions. The action can be executed sequentially if necessary.

        Args:
            action (np.ndarray): The action to execute, provided by the agent.

        Returns:
            Tuple[np.ndarray, Union[float, int], bool, bool, dict]: A tuple containing the following:
                - Observation vector as np.ndarray.
                - The accumulated reward as either float or int.
                - A boolean indicating if the episode is terminated.
                - A boolean indicating if the episode is truncated.
                - A dictionary containing additional environment info.
        """
        # For some reason sometime the environment is done but ray does not reinit it
        if self.single_env.done:
            logging.info("The Grid2Op Environment seems to be not initialized. We just return the last"
                         "observation and do not run anything")
            self.done = True
            self.truncated, self.terminated = self.check_terminated_truncated()

            return self.obs_rl, 0.0, self.terminated, self.truncated, self.info

        # Here we have the topology grid:
        if self.topology_actions:
            topo_vect = self.single_env.get_obs().topo_vect
            new_topology = self.actions[action]
            transf_act = set_bus_from_topo_vect(topo_vect, new_topology, self.single_env.action_space).to_vect()

        else:
            transf_act = self.actions[action]

        cur_obs, reward, self.done, self.info = split_and_execute_action(env=self.single_env,
                                                                         action_vect=transf_act)

        # get action:
        self.step_in_env += 1
        self.obs_grid2op = cur_obs

        # Select subset if wanted
        if isinstance(self.subset, list):
            obs_rl = self.obs_grid2op.to_vect()[self.subset]
        elif self.subset:
            obs_rl = obs_to_vect(self.obs_grid2op, False)
        else:
            obs_rl = self.obs_grid2op.to_vect()

        if self.scaler:
            if isinstance(self.scaler, BaseEstimator):
                obs_rl = self.scaler.transform(obs_rl.reshape(1, -1)).reshape(-1, )
            elif isinstance(self.scaler, ObjectRef):
                obs_rl = ray.get(self.scaler).transform(obs_rl.reshape(1, -1)).reshape(-1, )

        self.obs_rl = obs_rl

        if (cur_obs.rho.max() < self.action_threshold) and self.done is False:
            self.obs_rl = self.run_next_action()
            reward += self.passiv_reward


        self.done = self.single_env.done

        # Now check for Truncation
        self.truncated, self.terminated = self.check_terminated_truncated()

        # We add the passive reward to the overall reward and scale it !
        game_over = 0
        if self.truncated:
            game_over = -300
        if self.terminated:
            game_over = 500

        self.reward = reward

        return self.obs_rl, self.reward, self.terminated, self.truncated, self.info

    def reset(self, seed: Optional[int] = None, options: Optional[Dict[str, int]] = None) -> Tuple[np.ndarray, dict]:
        """
        Resets the environment to its initial state.

        This method resets the environment, optionally using a specific seed or resetting to a specific chronic
        ID provided in the options. It ensures the environment is ready for the agent's intervention.

        Args:
            seed (Optional[int]): An optional seed value for resetting the environment.
            options (Optional[Dict[str, int]]): Optional dictionary to reset to a specific chronic ID, e.g.,
                {"id": 42}.

        Returns:
            Tuple[np.ndarray, dict]: A tuple containing the following:
                - The observation vector after resetting the environment.
                - A dictionary with additional information about the environment.
        """
        if seed:
            self.single_env.seed(seed)
        if isinstance(options, dict):
            id = options.get("id", None)
            if id:
                self.single_env.set_id(id)

        self.done = True
        while self.done:
            # We have to iterate over the environment, because it can happen that the environment
            # can be finished without the intervention of the agent. By this loop we ensure that the
            # Senior is needed.
            self.obs_grid2op = self.single_env.reset()
            self.step_in_env = 0
            self.passiv_reward = 0
            self.obs_rl = self.run_next_action()
            if not self.single_env.done:
                self.done = False

        self.truncated, self.terminated = self.check_terminated_truncated()

        return self.obs_rl, self.info

    def render(self, mode='human') -> None:
        """
        Renders the environment (required by Gym).

        This method is a placeholder for rendering the environment, which is required by the Gym interface.
        No changes are made by this method.

        Args:
            mode (str): A string representing the rendering mode. Defaults to 'human'.

        Returns:
            None.
        """

    def check_terminated_truncated(self) -> Tuple[bool, bool]:
        """
        Checks whether the environment is terminated or truncated.

        This method checks if the current episode is terminated or truncated based on the environment's status.
        Termination happens when the maximum episode duration is reached, while truncation occurs when the environment
        is finished early.

        Returns:
            Tuple[bool, bool]: A tuple containing two booleans:
                - The first boolean indicates if the episode is truncated.
                - The second boolean indicates if the episode is terminated.
        """

        if self.done:
            trunc, term = True, False
            # Check for termination (and revert truncation)
            if self.single_env.nb_time_step == self.single_env.chronics_handler.max_episode_duration():
                trunc, term = False, True
        else:
            trunc, term = False, False

        return trunc, term



import logging
import pickle
from typing import Union, Optional, Tuple

import torch
import grid2op
import numpy as np
from pathlib import Path

from lightsim2grid import LightSimBackend
from sklearn.base import BaseEstimator
from torch.utils.data import Dataset


class Grid2OpDataset(Dataset):
    def __init__(
            self,
            root: Union[str, Path],
            dataset_name: str,
            env="l2rpn_case14_sandbox",
            split="train",
            scaler: Union[str, Path, BaseEstimator] = None
    ):
        """Constructor of the dataset representing grid2op networks as graphs

        Args:
            root: root folder where data is stored
            dataset_name: Name of the tutor results.
            num_actions: Number of actions. Necessary for the action space
            env: grid32op environment for which the graph data will be generated or String indicating name of env
            transform: PyG transform to be applied to the graph before passing it to model
            split: What dataset are we looking for?
            scaler: A scaler to scale the data prior to execution.
        """

        super(Grid2OpDataset).__init__()

        if type(env) == str:
            self.env = grid2op.make(env, backend=LightSimBackend())
        else:
            self.env = env

        self.dataset_name = dataset_name
        self.dataset_path = Path(root)
        self.split = split

        self.scaler = None
        if isinstance(scaler, BaseEstimator):
            self.scaler = scaler
        elif isinstance(scaler, (str, Path)):
            try:
                with open(scaler, "rb") as fp:  # Pickling
                    self.scaler = pickle.load(fp)
            except Exception as e:
                logging.info(f"The scaler provided could not be loaded by pickle")

        self.s, self.a = self.load_dataset()


    def load_dataset(self):
        """ Load the dataset from the given path. If a scaler was provided, we also scale the data

        Returns: Tuple with observation and action.

        """

        path = self.dataset_path / f"{self.dataset_name}_{self.split}.npz"
        data = np.load(path)

        if "val" in self.split:
            s_dat = data["s_validate"]
            a_dat = data["a_validate"]
        else:
            s_dat = data[f"s_{self.split}"]
            a_dat = data[f"a_{self.split}"]

        if self.scaler:
            s_dat = self.scaler.transform(s_dat)

        return (s_dat, a_dat)

    def process(self):
        pass

    def __len__(self):
        return len(self.a)

    def _download(self):
        pass

    def _process(self):
        pass

    def __getitem__(self,index:Optional[int]=None):
        """

        Args:
            idx (): index of sample to be retrieved from dataset
        Returns:
            a torch geometric Data object representing the grid2op grid corresponding to the index in the train data

        """
        if index is None:
            index = np.random.choice(len(self.a))

        state = torch.Tensor(self.s[index]).type(torch.float)
        action = torch.Tensor(self.a[index]).type(torch.long)
        return state, action



"""In this file, a neural network is developed to fit the dataset generated by Tutor.
Depending on the observation space and action space, the tutor model can/has to be
adjusted.

The Junior model returns a one-hot encoded output, based on the number of actions.

Credit: The junior is a more general approach of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""
import logging
import warnings
from collections import ChainMap, OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Tuple, List, Union, Optional, TypedDict

import grid2op.Environment
import lightning as L
import nni
import numpy as np
import torch
import torch.nn as nn
import torchmetrics
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.callbacks import EarlyStopping, ModelCheckpoint
from sklearn.base import BaseEstimator

from curriculumagent.common.obs_converter import obs_to_vect
from curriculumagent.common.utilities import map_actions
from curriculumagent.junior.torch_datasets import Grid2OpDataset

import nni

logging.basicConfig(level=logging.INFO)


class JuniorParamTorch(TypedDict):
    """
    A dictionary that defines the hyperparameters and paths required for the Junior Torch model.

    Attributes:
        action_space_path (Union[Path, List[Path]]): Either path to the actions or a list containing multiple actions.
        data_path (Path): Path of the Grid2Op environment.
        action_threshold (float): A threshold value between 0 and 1 that defines the action cutoff.
        subset (Union[bool, List[int]]): If True, filters the obs.to_vect based on predefined values. If False, all values
            are considered. Alternatively, one can provide their own subset of values.
        testing (bool): Indicator of whether the Grid2Op environment should be started in testing mode.
        activation (str): The activation function to be used in the model.
        learning_rate (Optional[float]): The learning rate for the model training. Defaults to None.
        layer1 (int): The size of the first layer.
        layer2 (int): The size of the second layer.
        layer3 (int): The size of the third layer.
        layer4 (int): The size of the fourth layer.
        batchsize (int): The batch size for training.
        dropout1 (float): Dropout probability for the first dropout layer.
        dropout2 (float): Dropout probability for the second dropout layer.
        epochs (int): The number of epochs for training.
        initializer (str): The initializer used for weight initialization.
    """

    activation: str
    learning_rate: Optional[float]
    layer1: int
    layer2: int
    layer3: int
    layer4: int
    batchsize: int
    dropout1: float
    dropout2: float
    epochs: int
    initializer: str


class JuniorTorch:
    """
    Defines the Junior Torch model. This class is responsible for configuring and managing the training and evaluation process of the Junior model
    based on PyTorch and PyTorch Lightning frameworks.
    """

    def __init__(
            self,
            action_space_file: Union[Path, List[Path]],
            config: JuniorParamTorch = None,
            seed: Optional[int] = None,
            run_nni: bool = False
    ):
        """ Initializes the Junior Torch model with the specified action space and configuration.

            Except setting all variables, the init additionally requires the size of the train set and optionally
            the number of epochs.

            Note:
                Pass epochs, learning_rate, batchsize, layer_size in the config.

            Args:
                action_space_file (Union[Path, List[Path]]): The action space file(s) used during training. This is required
                to define the input size of the Junior model.
                config (JuniorParamTorch, optional): Dictionary containing hyperparameters for the model. Defaults to None.
                seed (Optional[int]): Seed for reproducibility. Defaults to None.
                run_nni (bool): Flag indicating if NNI is used. If True, a callback is added. Defaults to False.

            Returns:
                None.
        """
        # Set the seed
        self.seed = seed
        if self.seed:
            np.random.seed(self.seed)
            torch.random.manual_seed(self.seed)

        # Callback:
        if run_nni:
            # ToDo Build NNi Callback.
            self.callback = [SendMetrics()] #ToDo: Need to be checked
        else:
            self.callback = []

        self.config = config
        self.batch_size = config.get("batchsize", 256)
        self.epochs = config.get("epochs", 1000)

        self.action_space_file = action_space_file
        self.lightning_model = Junior_PtL(action_space_file=action_space_file,
                                          model_config=self.config)

    def train(self,
              dataset_path: Path,
              target_model_path: Path,
              dataset_name: str = "junior_dataset",
              epochs: int = 1000,
              scaler: Optional[Union[str, Path, BaseEstimator]] = None,
              **kwargs
              ):
        """
        Trains the Junior Torch model.

        Args:
            dataset_path (Path): Path to the dataset files.
            target_model_path (Path): Path where the trained model will be saved.
            dataset_name (str): The name of the dataset file (e.g., `{dataset_name}_train.npz`). Defaults to "junior_dataset".
            epochs (int): The number of epochs to train. Defaults to 1000.
            scaler (Optional[Union[str, Path, BaseEstimator]]): Optional scaler for data normalization. Defaults to None.

        Returns:
            History: A record of the training process, including loss and accuracy.
        """

        # Create Datasets
        dat_tr = Grid2OpDataset(
            root=dataset_path,
            dataset_name=dataset_name,
            split="train",
            scaler=scaler)

        dat_val = Grid2OpDataset(
            root=dataset_path,
            dataset_name=dataset_name,
            split="val",
            scaler=scaler)

        dl_tr = torch.utils.data.DataLoader(dat_tr, batch_size=self.batch_size)
        dl_val = torch.utils.data.DataLoader(dat_val, batch_size=self.batch_size)
        if not epochs:
            epochs = self.epochs

        patience = kwargs.get("patience", False)
        history = History()
        callbacks = [history]
        if patience:
            early_stopping = EarlyStopping(monitor='val_loss', patience=patience)
            callbacks += [early_stopping]

        model_checkpoint = ModelCheckpoint(dirpath=target_model_path)
        callbacks += [model_checkpoint]

        # run model for dummy batch to initialize LazyLinear first layer and then initialize weights
        x, _ = next(iter(dl_tr))
        self.lightning_model.forward(x)
        self.lightning_model.apply_initializer(self.lightning_model.initializer)

        trainer = L.Trainer(max_epochs=epochs, callbacks=callbacks)
        trainer.fit(model=self.lightning_model, train_dataloaders=dl_tr, val_dataloaders=dl_val)

        return history

    def test(self,
             dataset_path: Path,
             dataset_name: str = "junior_dataset",
             checkpoint_path: Optional[Path] = None,
             scaler: Optional[Union[str, Path, BaseEstimator]] = None, ):
        """
        Tests the Junior Torch model on the specified dataset.

        Args:
            dataset_path (Path): Path to the dataset files.
            dataset_name (str): The name of the dataset file (e.g., `{dataset_name}_test.npz`). Defaults to "junior_dataset".
            checkpoint_path (Optional[Path]): Path to the model checkpoint for loading the trained model. Defaults to None.
            scaler (Optional[Union[str, Path, BaseEstimator]]): Optional scaler for data normalization. Defaults to None.

        Returns:
            dict: A dictionary containing the accuracy metrics for top-N predictions.
        """

        dat_test = Grid2OpDataset(
            root=dataset_path,
            dataset_name=dataset_name,
            #env=self.env,
            split="test",
            scaler=scaler)

        trainer = L.Trainer()

        if isinstance(checkpoint_path, Path):
            self.lightning_model = Junior_PtL.load_from_checkpoint(checkpoint_path=checkpoint_path)
            logging.info(f"Imported model from {checkpoint_path}")


        dl_test = torch.utils.data.DataLoader(dat_test.s, batch_size=1)

        a_pred = trainer.predict(self.lightning_model, dataloaders=dl_test)

        top_n = []

        for i in range(len(a_pred)):
            pred_numpy = a_pred[i].numpy()
            top_n.append(pred_numpy.argsort()[-20:])

        # Added accuracy to record the prediction performance
        accuracy = {}

        for n in range(1, 21):
            correct = 0
            for i in range(len(a_pred)):
                if dat_test.a[i, 0] in top_n[i][-n:]:
                    correct += 1

            acc = correct / len(a_pred) * 100
            logging.info(f"accuracy of top-{n} is {acc}")

            accuracy["accuracy of top-%d" % n] = correct / len(a_pred) * 100

        return accuracy


class Junior_PtL(L.LightningModule):
    """
        A PyTorch Lightning module representing the Junior model architecture
    """
    def __init__(self,
                 action_space_file: Union[Path, List[Path]] = None,
                 model_config: dict = {}
                 ):
        """
          Initializes the Junior_PtL model with the given action space and configuration.

          Args:
              action_space_file (Union[Path, List[Path]]): Path or list of paths to the action space files.
              model_config (dict): Configuration dictionary containing hyperparameters such as layer sizes, activation
                  functions, and initializers.

          Returns:
              None.
        """
        super().__init__()
        self.save_hyperparameters()
        list_of_actions = []
        if isinstance(action_space_file, Path):
            assert action_space_file.is_file()
            list_of_actions = [np.load(Path(action_space_file))]

        elif isinstance(action_space_file, list):
            for act_path in action_space_file:
                assert act_path.is_file()
            list_of_actions = [np.load(act_path) for act_path in action_space_file]
        self.num_actions = len(dict(ChainMap(*map_actions(list_of_actions))))

        self.accuracy = torchmetrics.classification.Accuracy(task="multiclass",
                                                             num_classes=self.num_actions)
        self.valaccuracy = torchmetrics.classification.MulticlassAccuracy(num_classes=self.num_actions)
        self.valaccuracy20 = torchmetrics.classification.MulticlassAccuracy(num_classes=self.num_actions,
                                                                            top_k=20)

        self.config = model_config
        self.lr = self.config.get("learning_rate", 5e-4)
        self.batch_size = self.config.get("batchsize", 256)
        self.epochs = self.config.get("epochs", 1000)

        self.layer_size, self.initializer, self.activation = self._extract_config(self.config)

        self.model = self._build_model()
        self.log_softmax = nn.LogSoftmax(dim=1)

    def training_step(self, batch, batch_idx):
        """
            Defines the training loop step. Performs a forward pass, computes the loss, and logs the accuracy.

            Args:
                batch (Tuple[Tensor, Tensor]): A batch of training data (inputs and targets).
                batch_idx (int): The index of the batch.

            Returns:
                Tensor: The computed loss value for the current batch.
        """
        # training_step defines the train loop.
        x, y = batch
        # forward to model
        pred = self.forward(x)
        loss = nn.NLLLoss()
        loss_res = loss(pred, y.reshape(-1))
        self.accuracy(pred, y.reshape(-1))
        self.log('loss', loss_res, on_step=False, on_epoch=True, prog_bar=True)
        self.log('accuracy', self.accuracy, on_step=False, on_epoch=True, prog_bar=True)
        return loss_res

    def configure_optimizers(self):
        """
        Configures the optimizer used for training. Adam optimizer is used with the specified learning rate.

        Returns:
            torch.optim.Optimizer: The optimizer for the model parameters.
        """
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer

    def validation_step(self, batch, batch_idx):
        """
        Defines the validation loop step. Performs a forward pass, computes the validation loss, and logs accuracy.

        Args:
            batch (Tuple[Tensor, Tensor]): A batch of validation data (inputs and targets).
            batch_idx (int): The index of the batch.

        Returns:
            Tensor: The computed validation loss value for the current batch.
        """
        # training_step defines the train loop.
        x, y = batch

        # forward to model
        pred = self.forward(x)
        loss = nn.NLLLoss()
        loss_res = loss(pred, y.reshape(-1))
        self.valaccuracy(pred, y.reshape(-1))
        self.valaccuracy20(pred, y.reshape(-1))

        self.log('val_loss', loss_res, on_step=False, on_epoch=True, prog_bar=True)
        self.log('val_accuracy', self.valaccuracy, on_step=False, on_epoch=True, prog_bar=True)
        self.log('val_accuracy_top20', self.valaccuracy20, on_step=False, on_epoch=True, prog_bar=True)
        return loss_res



    def _build_model(self) -> nn.Module:
        """
        Builds and returns the neural network model based on the provided configuration or default settings.

        Returns:
            nn.Module: The constructed PyTorch model as a sequential neural network.
        """

        layers = []

        if not self.config:
            layers.extend([
                ("layer1", nn.LazyLinear(out_features=1000)),  # Matches Senior's layer1
                ("activation1", self.activation),
                ("layer2", nn.Linear(in_features=1000, out_features=1000)),  # Matches Senior's layer2
                ("activation2", self.activation),
                ("layer3", nn.Linear(in_features=1000, out_features=1000)),  # Matches Senior's layer3
                ("activation3", self.activation),
                ("dropout1", nn.Dropout(0.25)),  # Dropout layers can be skipped in Senior
                ("layer4", nn.Linear(in_features=1000, out_features=1000)),  # Matches Senior's layer4
                ("activation4", self.activation),
                ("dropout2", nn.Dropout(0.25)),  # Dropout layers can be skipped in Senior
                ("act_layer", nn.Linear(in_features=1000, out_features=self.num_actions)),  # Matches Senior's act_layer
            ])
        else:
            # Build model based on hyperparameters:
            layers = [("layer1", nn.LazyLinear(out_features=self.layer_size[0])),
                      ("activation1", self.activation)]

            if self.config["layer2"] != 0:
                layers += [
                    ("layer2", nn.Linear(in_features=self.layer_size[1], out_features=self.layer_size[2])),
                    ("activation2", self.activation)
                ]

            if self.config["layer3"] != 0:
                layers += [
                    ("layer3", nn.Linear(in_features=self.layer_size[2], out_features=self.layer_size[3])),
                    ("activation3", self.activation)
                ]

            if self.config["dropout1"] != 0.0:
                layers += [("dropout1", nn.Dropout(self.config["dropout1"]))]

            if self.config["layer4"] != 0:
                layers += [
                    ("layer4", nn.Linear(in_features=self.layer_size[3], out_features=self.num_actions)),
                    ("activation4", self.activation)
                ]

            if self.config["dropout2"] != 0.0:
                layers += [("dropout2", nn.Dropout(self.config["dropout2"]))]

            layers.append(("act_layer", nn.Linear(in_features=self.layer_size[-1], out_features=self.num_actions)))

        model = nn.Sequential(OrderedDict(layers))  # Using OrderedDict to maintain the named layers
        return model

    def forward(self, x):
        """
        Forward pass of the Junior model.

        Args:
            x (Tensor): Input torch tensor for the forward pass.

        Returns:
            Tensor: The model's output tensor (represents action).
        """
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x)
        output = self.model(x)
        if output.dim() == 1:  # Handle 1D case
            action = self.log_softmax(output.unsqueeze(0)).squeeze(0)
        else:  # Handle 2D case (common case)
            action = self.log_softmax(output)
        return action

    def _extract_config(self, config) -> Tuple[List, any, any]:
        """
            Extracts model configuration such as layer sizes, activation function, and initializer from the provided config.

            Args:
                config (dict): Configuration dictionary with model parameters.

            Returns:
                Tuple[List[int], Callable, Callable]: A tuple containing the layer sizes, activation function, and initializer.
        """
        layer_size = [1000, 1000, 1000, 1000]
        initializer = nn.init.orthogonal_
        activation = nn.ReLU()

        if not any([k in config.keys() for k in
                    ["activation", "initializer", "layer1", "layer2", "layer3", "layer4"]]):
            warnings.warn("The custom dictionary did not have the correct keys. Using default model.")
            return layer_size, initializer, activation

        activation_option = {"leaky_relu": nn.LeakyReLU(inplace=True),
                             "relu": nn.ReLU(inplace=True),
                             "elu": nn.ELU(inplace=True)}
        initializer_option = {"O": nn.init.orthogonal_,
                              "RN": nn.init.normal_,
                              "RU": nn.init.uniform_,
                              "Z": nn.init.zeros_}

        if "activation" in config:
            activation = activation_option[config["activation"]]

        if "initializer" in config:
            initializer = initializer_option[config["initializer"]]

        if all([l in config for l in ["layer1", "layer2", "layer3", "layer4"]]):
            layer_size = [int(round(config[l])) for l in ["layer1", "layer2", "layer3", "layer4"]]

        return layer_size, initializer, activation

    def apply_initializer(self, initializer):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                initializer(m.weight)


class History(Callback):
    """
    A PyTorch Lightning callback to store the training and validation metrics at the end of each epoch.

    This callback collects metrics such as loss and accuracy during training and validation, and stores them in
    a dictionary for later use or analysis.

    """

    def __init__(self):
        """
        Initializes the History callback by creating an empty dictionary to store training and validation metrics.

        Returns:
            None.
        """
        self.history = {"loss": [],
                        "accuracy": [],
                        "val_loss": [],
                        "val_accuracy": [],
                        "val_accuracy_top20": []}

    def on_train_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """
        Called at the end of each training epoch to copy the metrics from the trainer and store them in the history.

        Args:
            trainer (pl.Trainer): The PyTorch Lightning trainer instance.
            pl_module (pl.LightningModule): The PyTorch Lightning module (model) being trained.

        Returns:
            None.
        """
        results = trainer.callback_metrics.copy()

        for k in self.history.keys():
            self.history[k].append(float(results[k]))


class SendMetrics(Callback):
    """
    A PyTorch Lightning callback to send training and validation metrics to the NNI framework.

    This callback is used to report metrics such as accuracy and loss at the end of each training epoch and after
    validation, enabling integration with the NNI (Neural Network Intelligence) framework for automated machine learning.
    """

    def on_train_epoch_end(self, trainer, pl_module):
        """
        Called at the end of each training epoch to send the training accuracy and loss to the NNI framework.

        Args:
            trainer (pl.Trainer): The PyTorch Lightning trainer instance.
            pl_module (pl.LightningModule): The PyTorch Lightning module (model) being trained.

        Returns:
            None.
        """
        # Get the training accuracy and loss from the logged metrics
        train_accuracy = trainer.callback_metrics.get("accuracy")
        train_loss = trainer.callback_metrics.get("loss")

        # Report metrics to NNI
        metrics = {}
        if train_accuracy is not None:
            metrics['train_accuracy'] = train_accuracy.item()

        if train_loss is not None:
            metrics['train_loss'] = train_loss.item()

        if metrics:
            nni.report_intermediate_result(metrics)

    def on_validation_end(self, trainer, pl_module):
        """
        Called at the end of the validation loop to send the validation accuracy and loss to the NNI framework.

        Args:
            trainer (pl.Trainer): The PyTorch Lightning trainer instance.
            pl_module (pl.LightningModule): The PyTorch Lightning module (model) being trained.

        Returns:
            None.
        """
        # Get the validation accuracy and loss from the logged metrics
        val_accuracy = trainer.callback_metrics.get("val_accuracy")
        val_loss = trainer.callback_metrics.get("val_loss")

        # Report metrics to NNI
        metrics = {}
        if val_accuracy is not None:
            metrics['val_accuracy'] = val_accuracy.item()

        if val_loss is not None:
            metrics['val_loss'] = val_loss.item()

        if metrics:
            nni.report_intermediate_result(metrics)




"""This file contains features/mappings/extractions for training."""
import logging
import warnings
from typing import Optional

import grid2op
import numpy as np

def obs_to_vect(obs: grid2op.Observation.BaseObservation, connectivity: bool = False) -> np.ndarray:
    """Method to convert only a subset of the observation to a vector.

    Args:
        obs: Original observation of Grid2Op.
        connectivity: Indicator, whether the connectivity matrix should be saved as well.

    Returns:
        Vector of the observation.

    """
    features = [
        # Timestamp Features
        [obs.month, obs.day, obs.hour_of_day, obs.minute_of_hour, obs.day_of_week],
        # Generation and Load
        obs.gen_p,
        obs.gen_q,
        obs.gen_v,
        obs.load_p,
        obs.load_q,
        obs.load_v,
        # Raw line values
        obs.p_or,
        obs.q_or,
        obs.v_or,
        obs.a_or,
        obs.p_ex,
        obs.q_ex,
        obs.v_ex,
        obs.a_ex,
        obs.rho,
        obs.line_status,
        obs.timestep_overflow,
        # Bus information
        obs.topo_vect,
        # cool downs:
        obs.time_before_cooldown_line,
        obs.time_before_cooldown_sub,
        # maintenance
        obs.time_next_maintenance,
        obs.duration_next_maintenance,
    ]

    if connectivity:
        features.append(obs.connectivity_matrix().reshape(-1))

    return np.concatenate(features, dtype=np.float32)


def vect_to_dict(
        vect: np.ndarray, examplary_obs: grid2op.Observation.BaseObservation, connectivity: bool = False
) -> dict:
    """Method that converts the vector of the obs_to_vect method to a dictionary.
    Note that  for this one we require an observation of the environment in order to gather the correct information.

    Args:
        vect: Vector of the obs subset
        examplary_obs: One Grid2Op environment to get the correct lengths.
        connectivity: Whether to return the connectivity or not. This is only possible, if
        connectivity matrix to begin with was saved before.

    Returns:
        A dictionary of the observation.

    """
    if not isinstance(vect, np.ndarray):
        raise TypeError("vect input does not have the correct type")
    if not isinstance(examplary_obs, grid2op.Observation.BaseObservation):
        raise TypeError("examplary_obs input does not have the correct type. Please enter the observation "
                        "of the grid2op environment")

    assert len(vect.shape) < 2, "The dimensions of the vect input are not correct. Should be a vector"

    out = {
        # The first 5 are allways the same:
        "month": vect[0],
        "day": vect[1],
        "hour_of_day": vect[2],
        "minute_of_hour": vect[3],
        "day_of_week": vect[4],
    }
    i = 5
    obs_json = examplary_obs.to_json()

    for k in [
        "gen_p",
        "gen_q",
        "gen_v",
        "load_p",
        "load_q",
        "load_v",
        "p_or",
        "q_or",
        "v_or",
        "a_or",
        "p_ex",
        "q_ex",
        "v_ex",
        "a_ex",
        "rho",
        "line_status",
        "timestep_overflow",
        "topo_vect",
        "time_before_cooldown_line",
        "time_before_cooldown_sub",
        "time_next_maintenance",
        "duration_next_maintenance",
    ]:
        out[k] = vect[i: i + len(obs_json[k])]
        i += len(obs_json[k])

    if connectivity:
        if np.sqrt(len(vect[i:])) % 1 == 0:
            c_m = vect[i:].reshape(examplary_obs.connectivity_matrix().shape)
            out["connectivity_matrix"] = c_m
        else:
            logging.warning("The connectivity Matrix is not quadratic. Thus, it is not added to the dictionary")

    return out



"""
This file is running through various seeds, to check the configurations fo different seed s
"""

import pickle

import matplotlib.pyplot as plt
from pathlib import Path

import numpy as np
from lightsim2grid import LightSimBackend
import grid2op
import sys
import logging
from typing import Optional, Tuple

from l2rpn_baselines.ExpertAgent.expertAgent import ExpertAgent
from grid2op.Agent import DoNothingAgent
from grid2op.Environment import BaseEnv
from curriculumagent.common.score_agent import load_or_run, render_report
from curriculumagent.tutor.tutors.general_tutor import GeneralTutor
from curriculumagent.submission.my_agent import MyAgent
from curriculumagent.tutor.tutors.n_minus_one_tutor import NminusOneTutor

plt.rcParams['figure.figsize'] = [15, 5]
date_strftime_format = "%Y-%m-%y %H:%M:%S"
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(asctime)s %(message)s",
                    datefmt=date_strftime_format)


def run_evaluation_of_env(agent_dict: dict, env: BaseEnv, out_path: Path, seed: Optional[int]) -> Tuple[dict, dict]:
    """ Calculate the performance of the provided agents (from the agents dicts)

    Note: The do nothing agent is always provided!

    Args:
        agent_dict: Dictionary containing multiple agents
        env: Grid2Op environment
        out_path: path, where to save the results and cached results
        seed: Optional seed for execution.

    Returns: None

    """
    if out_path.is_dir() is False:
        os.mkdir(out_path)

    number_of_runs = len(os.listdir(env.chronics_handler.path))

    do_nothing_agent = DoNothingAgent(env.action_space)
    # We use the reinit method to load 
    dn_report = load_or_run(agent=do_nothing_agent, env=env, output_path=out_path, name="DoNothing",
                            number_episodes=number_of_runs, seed=seed, reinit=True)  # Very Important, reinit True!!!
    print(f"The Do-Nothing agent has the scores of: {dn_report.score_data['all_scores']}")

    agent_res = []

    for name, agent in agent_dict.items():
        if ("Senior" in name) or ("Expert" in name):
            print("Run Senior Original with nb_process 1")
            agent_res.append(load_or_run(agent, env=env, output_path=out_path, name=name,
                                         nb_processes=1, number_episodes=number_of_runs, seed=seed))
        else:
            agent_res.append(load_or_run(agent, env=env, output_path=out_path, name=name,
                                         nb_processes=os.cpu_count(), number_episodes=number_of_runs, seed=seed))

    render_report(out_path / 'report.md', dn_report, agent_res)

    # collect the mean of the overall scores: 
    res_dict = {dn_report.agent_name: dn_report.avg_score}
    for agent in agent_res:
        res_dict[agent.agent_name] = agent.avg_score

    # Collect the surviving time
    surv_time = {dn_report.agent_name: dn_report.score_data['ts_survived']}
    for agent in agent_res:
        surv_time[agent.agent_name] = agent.score_data['ts_survived']

    return res_dict, surv_time


def create_agents_and_env():
    """ Simple Method to initialize the agents in order to make Pooling work

    Returns: dictionary of agent

    """
    backend = LightSimBackend()
    # Note we are only interested in the test env, thus number of episodes = 24
    env = grid2op.make("l2rpn_wcci_2022_valscenario", backend=backend)

    n_1_actions = [Path("/home/mlehna/AI2Go/l2rpn_binbinchen_iee/Update2023/wcci_2022_actions.npy"),
                   Path("/home/mlehna/AI2Go/l2rpn_binbinchen_iee/Update2023/maze_actions.npy")]
    disc_lines = [136, 146, 149, 156, 155, 180, 83, 154, 121, 93, 138, 88,
                  148, 86, 110, 106, 68, 162, 141, 79, 81, 150, 152, 153,
                  160, 131, 132, 126, 108, 62, 89, 61, 125, 91, 133, 82,
                  51, 175, 127, 84, 120, 76]

    tutor_n1 = NminusOneTutor(action_space=env.action_space,
                              action_space_file=n_1_actions,
                              do_nothing_threshold=0.9,
                              best_action_threshold=0.99,
                              rho_greedy_threshold=0.99,
                              lines_to_check=disc_lines,
                              return_status=True,
                              revert_to_original_topo=True)

    tutor_original = GeneralTutor(action_space=env.action_space,
                                  action_space_file=n_1_actions)

    agents = {"T Original": tutor_original,
              "T N-1": tutor_n1,
              }

    return agents, env


def run_experiment_with_seed(seed: int):
    """ Execute the experiment with the given seed

    Args:
        seed:

    Returns:

    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    fixed_path = Path("/share/data1/GYM/wcci2022/seeds") / str(seed)

    agents, env = create_agents_and_env()
    logging.info(f"Run evaluation of seed {seed}.")
    res = run_evaluation_of_env(agent_dict=agents,
                                env=env,
                                out_path=fixed_path,
                                seed=seed)
    logging.info(f"Seed {seed} done with score {res}")

    return res, seed


if __name__ == "__main__":
    """
    Create the different agents e.g. Tutors and Seniors
    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    import os

    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    np.random.seed(8888)
    seeds = list(np.random.randint(0, 10000, 10))
    logging.info(f"length of tasks:{len(seeds)}")

    #     # Save the results
    collect_scores = {}

    agents, env = create_agents_and_env()

    collect_survival_time = {}

    for seed in seeds:
        res, surv_time = run_evaluation_of_env(agent_dict=agents,
                                               env=env,
                                               out_path=Path("/share/data1/GYM/wcci2022/seeds") / str(seed),
                                               seed=seed)
        collect_scores[seed] = res
        collect_survival_time[seed] = surv_time

    with open('/share/data1/GYM/wcci2022/seeds/seed_res.pkl', 'wb') as handle:
        pickle.dump(collect_scores, handle)

    with open('/share/data1/GYM/wcci2022/seeds/collect_survival_time.pkl', 'wb') as handle:
        pickle.dump(collect_survival_time, handle)
